#
# Myokit symbolic expression classes. Defines different expressions, equations
# and the unit system.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import io
import math
import numpy

import myokit

from myokit import IntegrityError


# Expression precedence levels
FUNCTION_CALL = 70
POWER = 60
PREFIX = 50
PRODUCT = 40
SUM = 30
CONDITIONAL = 20
CONDITION_AND = 10
LITERAL = 0


class Expression:
    """
    Myokit's most generic interface for expressions. All expressions extend
    this class.

    Expressions are immutable objects.

    Expression objects have an ``_rbp`` property determining their *right
    binding power* (myokit uses a top-down operator precedence parsing scheme)
    and an optional ``_token`` property that may contain the text token this
    expression object was originally parsed from.

    *Abstract class*
    """
    _rbp = None     # Right-binding power (see parser).
    _rep = ''       # Mmt representation
    _treeDent = 2   # Tab size used when displaying as tree.

    def __init__(self, operands=None):
        self._token = None

        # Store operands
        self._operands = () if operands is None else operands

        # Store references
        self._references = set()
        for op in self._operands:
            self._references |= op._references

        # Contains partial derivatives or initial values?
        self._has_partials = False
        self._has_initials = False
        for op in self._operands:
            if op._has_partials:
                self._has_partials = True
            if op._has_initials:
                self._has_initials = True

        # Cached results:
        # Since expressions are immutable, the results of many methods can be
        # cached. These could be evaluated initially, but it's more efficient
        # to way until a first request.
        self._cached_hash = None
        self._cached_polish = None
        self._cached_validation = None
        self._cached_unit_tolerant = None
        self._cached_unit_strict = None

    def __bool__(self):
        # Determines the outcome of "if expression".
        return True

    def bracket(self, op=None):
        """
        Checks if the given operand (which should be an operand of this
        expression) needs brackets around it when writing a mathematical
        expression.

        For example, ``5 + 3`` will require a bracket when used in a
        multiplication (e.g. ``2 * (5 + 3)``), so calling ``bracket(5 + 3)`` on
        a multiplication will return ``True``.

        Alternatively, when used in a function the expression will not require
        brackets, as the function (e.g. ``sin()``) already provides them, so
        calling ``bracket(5 + 3)`` on a function will return False.
        """
        raise NotImplementedError

    def clone(self, subst=None, expand=False, retain=None):
        """
        Clones this expression.

        The optional argument ``subst`` can be used to pass a dictionary
        mapping expressions to a substitute. Any expression that finds itself
        listed as a key in the dictionary will return this replacement value
        instead.

        The argument ``expand`` can be set to True to expand all variables
        other than states. For example, if ``x = 5 + 3 * y`` and
        ``y = z + sqrt(2)`` and ``dot(z) = ...``, cloning x while expanding
        will yield ``x = 5 + 3 * (z + sqrt(2))``. When expanding, all constants
        are converted to numerical values. To maintain some of the constants,
        pass in a list of variables or variable names as ``retain``.

        Substitution takes precedence over expansion: A call such as
        ``e.clone(subst={x:y}, expand=True) will replace ``x`` by ``y`` but not
        expand any names appearing in ``y``.
        """
        raise NotImplementedError

    def code(self, component=None):
        """
        Returns this expression formatted in ``mmt`` syntax.

        When :class:`Name` objects are encountered, the fullly qualified name
        of the :class:`myokit.Variable` that they refer to is rendered, with
        the following exceptions:

        - if the variable's component matches the argument ``component`` or the
          variable is nested, then the local variable name is used
        - if the given ``component`` has an alias for the variable, this alias
          is used.
        """
        # Note: Because variable and component names can change, the output of
        # code can not be cached (for non-literal expressions).
        b = io.StringIO()
        self._code(b, component)
        return b.getvalue()

    def _code(self, b, c):
        """
        Internal version of ``code()``, should write the generated code to the
        stringbuffer ``b``, from the context of component ``c``.
        """
        raise NotImplementedError

    def __contains__(self, key):
        return key in self._operands

    def contains_type(self, kind):
        """
        Returns True if this expression tree contains an expression of the
        given type.
        """
        if isinstance(self, kind):
            return True
        if kind == PartialDerivative:
            return self._has_partials
        if kind == InitialValue:
            return self._has_initials
        for op in self:
            if op.contains_type(kind):
                return True
        return False

    def depends_on(self, lhs, deep=False):
        """
        Returns ``True`` if this :class:`Expression` depends on the given
        :class:`LhsExpresion`.

        With ``deep=False`` (default), only dependencies appearing directly in
        the expression are checked. With ``deep=True`` the method also checks
        the right-hand side equation defined in the model for any
        :class:`Name` or :class:`Derivative` it encounters.
        """
        # Shallow check
        if not deep:
            return lhs in self._references

        # Determine if this variable has a (deep) dependency on the given lhs
        done = set()
        if isinstance(self, myokit.LhsExpression):
            done.add(self)
        todo = set(self._references)
        while todo:
            ref = todo.pop()
            done.add(ref)
            if ref == lhs:
                return True
            elif ref._has_partials or ref._has_initials:
                # Partial derivatives and initial values have no rhs
                continue
            elif not ref._proper:
                # Values that are not variables count as independent
                continue

            var = ref.var()
            if ref.is_state_value() or var.is_bound():
                # State values and bound variables have no rhs
                continue
            else:
                todo.update(var.rhs()._references - done)
        return False

    def depends_on_state(self, deep=False):
        """
        Returns ``True`` if this :class:`Expression` depends on one or more
        state variables.

        With ``deep=False`` (default), only dependencies appearing directly in
        the expression are checked. With ``deep=True`` the method also checks
        the right-hand side equation defined in the model for any
        :class:`Name` or :class:`Derivative` it encounters.
        """
        # Shallow check
        if not deep:
            for lhs in self._references:
                if lhs.is_state_value():
                    return True
            return False

        # Determine if this variable has a (deep) dependency on a state
        todo = set(self._references)
        done = set()
        if isinstance(self, myokit.LhsExpression):
            done.add(self)
        while todo:
            ref = todo.pop()
            done.add(ref)
            if ref.is_state_value():
                return True
            elif ref._has_partials or ref._has_initials:
                # Partial derivatives and initial values have no rhs
                continue
            elif not ref._proper:
                # Values that are not variables count as independent
                continue
            var = ref.var()
            if not var.is_bound():
                todo.update(var.rhs()._references - done)
        return False

    def diff(self, lhs, independent_states=True):
        """
        Returns an expression representing the partial derivative of this
        expression with respect to the expression ``lhs``.

        The argument ``lhs`` must be a :class:`Name` or a
        :class:`InitialValue`, taking derivatives with respect to
        a class:`Derivative` or :class:`PartialDerivative` is not supported.

        **Expressions involving variables**

        Partial derivatives are determined recursively. If, at any point in
        this recursion, a :class:`Name` or :class:`Derivative` is encountered,
        this is handled in the following way:

        - The partial derivative of any :class:`Name` with respect to an
          identical ``Name`` is 1 (without units / dimensionless).
        - The partial derivative of a :class:`Name` referencing a state
          variable is zero if ``independent_states=True``, but will otherwise
          be represented as a :class:`PartialDerivative`.
        - The partial derivative of a :class:`Derivative` or of a :class:`Name`
          referencing a non-state variable, will both be determined based on
          the corresponding right-hand side expression. If this references the
          ``lhs``, then a ``PartialDerivative`` will be returned. If it does
          not reference the ``lhs`` and ``independent_states=True``, then zero
          will be returned. If it does not reference the ``lhs``, but
          ``independent_states=False`` and one or more states are referenced,
          then a :class:`PartialDerivative` will be returned.
        - The partial derivative of a :class:`Name` referencing a bound
          variable is zero.

        **Simplification**

        Some effort is made to eliminate expressions that evaluate to zero, but
        no further simplification is performed. Multiplications by 1 are
        preserved as these can provide valuable unit information.

        **Conditional expressions**

        When calculating derivatives, the following simplifying assumptions are
        made with respect to conditional expressions:

        - When evaluating conditional expressions (:class:`If` and
          :class:`Piecewise`), the discontinuities at the condition boundaries
          are ignored. Instead, the method simply returns a similar conditional
          expression with the original operands replaced by their derivatives,
          e.g. ``if(condition, a, b)`` becomes ``if(condition, a', b')``.

        **Discontinuous functions**

        Similarly, some functions are discontinuous so that their derivatives
        are undefined at certain points, but this method returns the
        right-derivative for those points. In particular:

        - The true derivative of ``floor(x)`` is zero when ``x`` is not an
          integer and undefined if it is, but this method always returns zero.
        - Similarly, the true derivative of ``ceil(x)`` is undefined when ``x``
          is an integer, but this method always returns zero.
        - The true derivative of an integer division ``a // b`` with respect to
          either ``a`` or ``b`` is zero when ``a`` is not an integer multiple
          of ``b``, and otherwise undefined; but this method always returns
          zero.
        - The true derivative of the remainder operation
          ``a(x) % b(x) = a - b * floor(a / b)`` is
          ``da/dx - db/dx * floor(a/b) - b * d/dx floor(a/b)``, but (as above)
          this method assumes the derivative of the floor function is zero, and
          so will always return ``da/dx - db/dx * floor(a/b)``.
        - The true derivative of ``abs(f(x))`` is f'(x) for ``x > 0``, -f'(x)
          for ``x < 0``, and undefined for ``x == 0``, but this method will
          return ``f'(x)`` for ``x >= 0`` and ``-f'(x)``s for ``x < 0``.

        **Non-integer powers**

        Since ``a(x)^b(x)`` is undefined for non-integer ``b(x)`` when
        `a(x) < 0``, the derivative of ``a(x)^b(x)`` is only defined if
        ``a(x) >= 0`` or ``b'(x) = 0``. No errors or warnings are given if the
        derivative is undefined, until the equations are evaluated (note that
        at this point evaluation of ``a(x)^b(x)`` will also fail).

        """
        # Check LHS
        if not isinstance(lhs, (Name, InitialValue)):
            raise ValueError(
                'Partial derivatives can only be taken with respect to a'
                ' myokit.Name or myokit.InitialValue.')

        # Get derivative or None for 0
        derivative = self._diff(lhs, bool(independent_states))

        # Result zero? Then ensure it has the right unit
        if derivative is None:
            derivative = Number(0, self._diff_unit(lhs))

        return derivative

    def _diff(self, lhs, idstates):
        """
        Internal version of ``diff()``.

        Assumes lhs is a Name; and will return None if this expression does not
        depend on ``lhs``.
        """
        raise NotImplementedError

    def _diff_unit(self, lhs):
        """
        Returns the unit that the derivative of this expression w.r.t. the
        given ``lhs`` _should_ be in.

        This is used by both :meth:`diff()` and several implementations of
        :meth:`_diff()`.
        """
        try:
            unit1 = self.eval_unit(myokit.UNIT_TOLERANT)
        except myokit.IncompatibleUnitError as e:
            unit1 = None
        unit2 = lhs.var().unit()

        if unit1 is None and unit2 is None:
            return None
        if unit1 is None:
            return 1 / unit2
        if unit2 is None:
            return unit1
        return unit1 / unit2

    def __eq__(self, other):
        # Equality checking method, used by all expressions.
        if self is other:
            return True
        elif type(self) != type(other):
            return False
        else:
            # Compare cached polish expression (which uses ids)
            # Note that the polish representation uses object ids instead of
            # qnames
            return self._polish() == other._polish()

    def eval(self, subst=None, precision=myokit.DOUBLE_PRECISION):
        """
        Evaluates this expression and returns the result.

        The optional argument ``subst`` can be used to pass a dictionary
        mapping :class:`LhsExpression` objects to expressions or numbers to
        substitute them with.

        For debugging purposes, the argument ``precision`` can be set to
        ``myokit.SINGLE_PRECISION`` to perform the evaluation with 32 bit
        floating point numbers.

        Note: This operation will fail if the expression contains :class:`Name`
        objects with a value other than an :class:`Expression` or (an object
        that can be cast to) a float.
        """
        # Check subst dict
        if subst:
            try:
                subst = dict(subst)
            except TypeError:
                raise ValueError('Argument `subst` must be dict or None.')
            for k, v in subst.items():
                if not isinstance(k, myokit.LhsExpression):
                    raise ValueError(
                        'All keys in `subst` must LhsExpression objects.')
                if not isinstance(v, myokit.Expression):
                    try:
                        v = float(v)
                    except ValueError:
                        raise ValueError(
                            'All values in `subst` must be Expression objects'
                            ' or numbers.')
                    subst[k] = myokit.Number(v)

        # Evaluate
        try:
            return self._eval(subst, precision)
        except EvalError as e:

            # It went wrong! Create a nice error message.
            out = [_expr_error_message(self, e)]

            # Show values of operands
            ops = [op for op in e.expr]
            if ops:
                out.append('With the following operands:')
                for i, op in enumerate(ops):
                    pre = '  (' + str(1 + i) + ') '
                    try:
                        out.append(pre + str(op._eval(subst, precision)))
                    except EvalError:
                        out.append(pre + 'another error')

            # Show variables included in expression
            refs = e.expr.references()
            if refs:
                out.append('And the following variables:')
                for ref in refs:
                    name = str(ref)

                    # Perform substitution, or get expression from lhs
                    if subst and ref in subst:
                        rhs = subst[ref]
                    else:
                        rhs = ref.rhs()

                    if isinstance(rhs, Number):
                        # Show numbers on same line
                        pre = '  ' + name + ' = '
                    else:
                        # Show expressions + results on next line
                        out.append('  ' + name + ' = ' + rhs.code())
                        pre = '  ' + ' ' * len(name) + ' = '

                    try:
                        out.append(pre + str(rhs._eval(subst, precision)))
                    except EvalError:
                        out.append(pre + 'another error')

            # Raise new exception with better message
            out = '\n'.join(out)
            raise myokit.NumericalError(out)

    def _eval(self, subst, precision):
        """
        Internal, error-handling free version of eval.
        """
        raise NotImplementedError

    def eval_unit(self, mode=myokit.UNIT_TOLERANT):
        """
        Evaluates the unit this expression should have, based on the units of
        its variables and literals.

        Incompatible units may result in a
        :class:`myokit.IncompatibleUnitError` being raised. The method for
        dealing with unspecified units can be set using the ``mode`` argument.

        Using ``myokit.UNIT_STRICT`` any unspecified unit will be treated as
        dimensionless. For example adding ``None + [kg]`` will be treated as
        ``[1] + [kg]`` which will raise an error. Similarly, ``None * [m]``
        will be taken to mean ``[1] * [m]`` which is valid, and ``None * None``
        will return ``[1]``. In strict mode, functions such as ``exp()``, which
        expect dimensionless input will raise an error if given a
        non-dimensionless operator.

        Using ``myokit.UNIT_TOLERANT`` unspecified units will try to be ignored
        where possible. For example, the expression ``None + [kg]`` will be
        treated as ``[kg] + [kg]``, while ``None * [m]`` will be read as
        ``[1] * [m]`` and ``None * None`` will return ``None``. Functions such
        as ``exp()`` will not raise an error, but simply return a dimensionless
        value.

        The method is intended to be used to check the units in a model, so
        every branch of an expression is evaluated, even if it won't affect the
        final result.

        In strict mode, this method will always either return a
        :class:`myokit.Unit` or raise an :class:`myokit.IncompatibleUnitError`.
        In tolerant mode, ``None`` will be returned if the units are unknown.
        """
        # Get cached unit or error
        if mode == myokit.UNIT_STRICT:
            result = self._cached_unit_strict
        else:
            result = self._cached_unit_tolerant

        # Evaluate and cache
        if result is None:
            try:
                result = self._eval_unit(mode)
            except EvalUnitError as e:
                result = myokit.IncompatibleUnitError(
                    _expr_error_message(self, e),
                    e.expr._token
                )

            if mode == myokit.UNIT_STRICT:
                self._unit_strict = result
            else:
                self._unit_tolerant = result

        # Raise error or return
        if isinstance(result, myokit.IncompatibleUnitError):
            raise result
        return result

    def _eval_unit(self, mode):
        """ Internal version of eval_unit(). """
        raise NotImplementedError

    def __float__(self):
        return float(self.eval())

    def __getitem__(self, key):
        return self._operands[key]

    def __hash__(self):
        if self._cached_hash is None:
            self._cached_hash = hash(self._polish())
        return self._cached_hash
        # Note: anything that has an __eq__ stops inheriting this hash method!
        # From: https://docs.python.org/3.1/reference/datamodel.html
        # > If a class that overrides __eq__() needs to retain the
        #   implementation of __hash__() from a parent class, the interpreter
        #   must be told this explicitly by setting
        #   __hash__ = <ParentClass>.__hash__. Otherwise the inheritance of
        #   __hash__() will be blocked, just as if __hash__ had been explicitly
        #   set to None.

    def __int__(self):
        return int(self.eval())

    def is_conditional(self):
        """
        Returns True if and only if this expression's tree contains a
        conditional statement.
        """
        for e in self:
            if e.is_conditional():
                return True
        return False

    def is_constant(self):
        """
        Returns true if this expression contains no references or only
        references to variables with a constant value.
        """
        for ref in self._references:
            if not ref.is_constant():
                return False
        return True

    def is_derivative(self, var=None):
        """
        Returns ``True`` only if this is a time-derivative, i.e. a
        :class:`myokit.Derivative` instance (and references the variable
        ``var``, if given).
        """
        return False

    def is_literal(self):
        """
        Returns ``True`` if this expression does not contain any references.
        """
        return len(self._references) == 0

    def is_name(self, var=None):
        """
        Returns ``True`` only if this expression is a :class:`myokit.Name`
        (and references the variable ``var``, if given).
        """
        return False

    def is_number(self, value=None):
        """
        Returns ``True`` only if this expression is a :class:`myokit.Number`.

        If the optional argument ``value`` is set, it will also return
        ``False`` if the number does not have the given value.
        """
        return False

    def is_state_value(self):
        """
        Returns ``True`` if this expression is a :class:`Name` pointing to the
        current value of a state variable.
        """
        return False

    def __iter__(self):
        return iter(self._operands)

    def __len__(self):
        return len(self._operands)

    def __ne__(self, other):
        return not self.__eq__(other)

    def operator_rep(self):
        """
        Returns a representation of this expression's type. (For example '+' or
        '*')
        """
        return self._rep

    def _polish(self):
        """
        Returns a reverse-polish notation version of this expression's code,
        using Variable id's instead of Variable name's to create immutable,
        unambiguous expressions.

        Note that id's are immutable _during the Variable's lifetime_. Since a
        Name object stores a reference to the Variable, this means the
        variable id is immutable in the expression's lifetime.
        """
        if self._cached_polish is None:
            b = io.StringIO()
            self._polishb(b)
            self._cached_polish = b.getvalue()
        return self._cached_polish

    def _polishb(self, b):
        """
        Recursive part of _polish(). Should write the generated code to the
        stringbuffer b.

        Because hash() and others use _polish(), this function should never
        use eval() or other advanced functions.
        """
        raise NotImplementedError

    def pyfunc(self, use_numpy=True):
        """
        Converts this expression to Python and returns the new function's
        handle.

        By default, when converting mathematical functions such as ``log``, the
        version from ``numpy`` (i.e. ``numpy.log``) is used. To use the
        built-in ``math`` module instead, set ``use_numpy=False``.
        """
        # Get expression writer
        if use_numpy:
            w = myokit.numpy_writer()
        else:
            w = myokit.python_writer()

        # Create function text
        args = [w.ex(x) for x in self._references]
        c = 'def ex_pyfunc_generated(' + ','.join(args) + '):\n    return ' \
            + w.ex(self)

        # Create function
        local = {}
        if use_numpy:
            exec(c, {'numpy': numpy}, local)
        else:
            exec(c, {'math': math}, local)

        # Return
        return local['ex_pyfunc_generated']

    def pystr(self, use_numpy=False):
        """
        Returns a string representing this expression in Python syntax.

        By default, built-in functions such as 'exp' are converted to the
        Python version 'math.exp'. To use the numpy versions, set
        ``numpy=True``.
        """
        # Get expression writer
        if use_numpy:
            w = myokit.numpy_writer()
        else:
            w = myokit.python_writer()

        # Return string
        return w.ex(self)

    def __reduce__(self):
        """ Called when attempting to pickle an expression. """
        raise NotImplementedError(
            'Individual myokit Expressions can not be pickled. Please try e.g.'
            ' pickling a full model, or pickling the output of'
            ' `Expression.code()` and following unpickling with a call to'
            ' `myokit.parse_expression(unpickled_code, context=a_model)` to'
            ' recreate the Expression.')

    def references(self):
        """
        Returns a set containing all references to variables made in this
        expression.
        """
        return set(self._references)

    def __repr__(self):
        return 'myokit.Expression[' + self.code() + ']'

    def __str__(self):
        return self.code()

    def tree_str(self):
        """
        Returns a string representing the parse tree corresponding to this
        expression.
        """
        b = io.StringIO()
        self._tree_str(b, 0)
        return b.getvalue()

    def _tree_str(self, b, n):
        raise NotImplementedError

    def validate(self):
        """
        Validates operands, checks cycles without following references. Will
        raise exceptions if errors are found.
        """
        return self._validate([])

    def _validate(self, trail):
        """
        The argument ``trail`` is used to check for cyclical refererences.
        """
        if self._cached_validation:
            return

        # Check for cyclical dependency
        if id(self) in trail:
            raise IntegrityError('Cyclical expression found', self._token)
        trail2 = trail + [id(self)]

        # It's okay to do this check with id's. Even if there are multiple
        # objects that are equal, if they're cyclical you'll get back round to
        # the same ones eventually. Doing this with the value requires hash()
        # which requires code() which may not be safe to use before the
        # expressions have been validated.
        for op in self:
            if not isinstance(op, Expression):
                raise IntegrityError(
                    'Expression operands must be other Expression objects.'
                    ' Found: ' + str(type(op)) + '.', self._token)
            op._validate(trail2)

        # Cache validation status
        self._cached_validation = True

    def walk(self, allowed_types=None):
        """
        Returns an iterator over this expression tree (depth-first). This is a
        slow operation. Do _not_ use in performance sensitive code!

        Example::

            5 + (2 * sqrt(x))

            1) Plus
            2) Number(5)
            3) Multiply
            4) Number(2)
            5) Sqrt
            6) Name(x)

        To return only expressions of certain types, pass in a sequence
        ``allowed_typess``, containing all types desired in the output.
        """
        if allowed_types is None:

            def walker(op):
                yield op
                for kid in op:
                    for x in walker(kid):
                        yield x
            return walker(self)

        else:

            if type(allowed_types) == type:
                allowed_types = set([allowed_types])
            else:
                allowed_types = set(allowed_types)

            def walker(op):
                if type(op) in allowed_types:
                    yield op
                for kid in op:
                    for x in walker(kid):
                        yield x
            return walker(self)


class Number(Expression):
    """
    Represents a number with an optional unit for use in Myokit expressions.
    All numbers used in Myokit expressions are floating point.

    >>> import myokit
    >>> x = myokit.Number(10)
    >>> print(x)
    10

    >>> x = myokit.Number(5.00, myokit.units.V)
    >>> print(x)
    5 [V]

    Arguments:

    ``value``
        A numerical value (something that can be converted to a ``float``).
        Number objects are immutable so no clone constructor is provided.
    ``unit``
        A unit to associate with this number. If no unit is specified the
        number's unit will be left undefined.

    *Extends:* :class:`Expression`
    """
    _rbp = LITERAL

    def __init__(self, value, unit=None):
        super().__init__()
        if isinstance(value, myokit.Quantity):
            # Conversion from Quantity class
            if unit is not None:
                raise ValueError(
                    'myokit.Number created from a myokit.Quantity cannot'
                    ' specify an additional unit.')
            self._value = value.value()
            self._unit = value.unit()
        else:
            # Basic creation with number and unit
            self._value = float(value) if value else 0.0
            if unit is None or isinstance(unit, myokit.Unit):
                self._unit = unit
            elif isinstance(unit, str):
                self._unit = myokit.parse_unit(unit)
            else:
                raise ValueError(
                    'Unit in myokit.Number should be a myokit.Unit or None.')
        # Create nice string representation
        self._str = myokit.float.str(self._value)
        if self._str[-2:] == '.0':
            # Turn 5.0 into 5
            self._str = self._str[:-2]
        elif self._str[-4:-1] == 'e-0':
            # Turn 1e-05 into 1e-5
            self._str = self._str[:-2] + self._str[-1]
        elif self._str[-4:-2] == 'e+':
            if self._str[-2:] == '00':
                # Turn 1e+00 into 1
                self._str = self._str[:-4]
            elif self._str[-2:-1] == '0':
                # Turn e+05 into e5
                self._str = self._str[:-3] + self._str[-1:]
            else:
                # Turn e+15 into e15
                self._str = self._str[:-3] + self._str[-2:]
        if self._unit and self._unit != myokit.units.dimensionless:
            self._str += ' ' + str(self._unit)

    def bracket(self, op=None):
        """See :meth:`Expression.bracket()`."""
        if op is not None:
            raise ValueError('Given operand is not used in this expression.')
        return False

    def clone(self, subst=None, expand=False, retain=None):
        """See :meth:`Expression.clone()`."""
        if subst and self in subst:
            return subst[self]
        return Number(self._value, self._unit)

    def _code(self, b, c):
        b.write(self._str)

    def convert(self, unit):
        """
        Returns a copy of this number in a different unit. If the two units are
        not compatible a :class:`myokit.IncompatibleUnitError` is raised.
        """
        return Number(myokit.Unit.convert(self._value, self._unit, unit), unit)

    def _diff(self, lhs, idstates):
        return None

    def _eval(self, subst, precision):
        if precision == myokit.SINGLE_PRECISION:
            return numpy.float32(self._value)
        else:
            return self._value

    def _eval_unit(self, mode):
        if mode == myokit.UNIT_STRICT and self._unit is None:
            return myokit.units.dimensionless
        return self._unit

    def is_constant(self):
        """See :meth:`Expression.is_constant()`."""
        return True

    def is_literal(self):
        """See :meth:`Expression.is_literal()`."""
        return True

    def is_number(self, value=None):
        """See :meth:`Expression.is_number()`."""
        return (value is None) or (value == self._value)

    def _polishb(self, b):
        b.write(self._str)

    def _tree_str(self, b, n):
        b.write(' ' * n + self._str + '\n')

    def unit(self):
        """
        Returns the unit associated with this Number or ``None`` if no unit was
        specified.
        """
        return self._unit

    def value(self):
        """Returns the value of this number."""
        return self._value


class LhsExpression(Expression):
    """
    An expression referring to the left-hand side of an equation.

    Running :func:`eval() <Expression.eval>` on an `LhsExpression` returns the
    evaluation of the associated right-hand side. This may result in errors if
    no right-hand side is defined. In other words, this will only work if the
    expression is embedded in a variable's defining equation.

    *Abstract class, extends:* :class:`Expression`
    """
    def _eval(self, subst, precision):
        if subst and self in subst:
            return subst[self].eval()
        try:
            return self.rhs()._eval(subst, precision)
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def is_constant(self):
        """See :meth:`Expression.is_constant()`."""
        return self.var().is_constant()

    def is_literal(self):
        """See :meth:`Expression.is_literal()`."""
        return False

    def rhs(self):
        """
        Returns the RHS expression equal to this LHS expression in the
        associated model.
        """
        raise NotImplementedError

    def var(self):
        """
        Returns the variable referenced by this `LhsExpression`.

        For :class:`Name` objects this will be equal to the left hand side of
        their defining equation, for :class:`Derivative` objects this will be
        the variable they represent the derivative of.
        """
        raise NotImplementedError


class Name(LhsExpression):
    """
    Represents a reference to a variable.

    *Extends:* :class:`LhsExpression`
    """
    _rbp = LITERAL
    __hash__ = LhsExpression.__hash__

    def __init__(self, value):
        super().__init__()
        self._value = value
        self._references = set([self])
        self._proper = isinstance(self._value, myokit.Variable)

    def bracket(self, op=None):
        """See :meth:`Expression.bracket()`."""
        if op is not None:
            raise ValueError('Given operand is not used in this expression.')
        return False

    def clone(self, subst=None, expand=False, retain=None):
        """See :meth:`Expression.clone()`."""
        if subst and self in subst:
            return subst[self]
        if expand and self._proper:
            if not self._value.is_state():
                if (retain is None) or (
                        self not in retain
                        and self._value not in retain
                        and self._value.qname() not in retain
                ):
                    return self._value.rhs().clone(subst, expand, retain)
        return Name(self._value)

    def _code(self, b, c):
        if self._proper:
            # Handle proper variable references
            if self._value.is_nested():
                b.write(self._value.name())
            else:
                if c:
                    try:
                        b.write(c.alias_for(self._value))
                        return
                    except KeyError:
                        pass
                b.write(self._value.qname(c))
        elif isinstance(self._value, str):
            # Allow strings for debugging
            b.write('str:' + str(self._value))
        else:
            # Allow "misusing" the expression system by storing other types as
            # values.
            b.write(str(self._value))

    def _diff(self, lhs, idstates):

        # Derivative w.r.t. self is one
        if lhs == self:
            return Number(1)

        # Value isn't a variable? Then always return an object
        if not self._proper:
            return PartialDerivative(self, lhs)

        # If idstates=False, a state variable reference returns an object, and
        # any state-dependent variable is dependent on the lhs.
        if not idstates:
            if self._value.is_state():
                return PartialDerivative(self, lhs)
            elif self._value.is_intermediary():
                # Possible state dependency: check!
                if self.depends_on_state(deep=True):
                    return PartialDerivative(self, lhs)

        # Otherwise, check for dependency (includes checks on this variable!)
        if self.depends_on(lhs, deep=True):
            return PartialDerivative(self, lhs)
        return None

    def _eval_unit(self, mode):

        if self._proper:
            # Note: Don't get it from the variable's RHS!
            # If the variable unit isn't specified:
            #  1. In tolerant mode a None will propagate without errors
            #  2. In strict mode, it is dimensionless (and if the RHS thinks
            #     otherwise the RHS is wrong).
            # In addition, for e.g. derivatives this can lead to cycles.
            return self._value.unit(mode)

        # Improper name, return dimensionless or None
        elif mode == myokit.UNIT_STRICT:
            return myokit.units.dimensionless
        return None

    def __eq__(self, other):
        if type(other) != Name:
            return False
        if self._proper:
            # Value is name? Then just check with is
            return self._value is other._value
        else:
            # Debug thing? Then convert to string and see if the
            # representations match. (This is the same as what would happen if
            # __eq__ was called on an expression _containing_ a Name, e.g
            # myokit.PrefixPlus(myokit.Name(1.23)).
            return self.code() == other.code()

    def is_name(self, var=None):
        """See :meth:`Expression.is_name()`."""
        return (var is None) or (var == self._value)

    def is_state_value(self):
        """See: meth:`Expression.is_state_value()`."""
        return self._proper and self._value.is_state()

    def _polishb(self, b):
        if isinstance(self._value, str):
            # Allow an exception for strings
            b.write('str:')
            b.write(self._value)
        else:
            # Use object id here to make immutable references. This is fine
            # since references should always be to variable objects, and
            # variables are unique this way (one Variable('x') doesn't equal
            # another Variable('x')).
            b.write('var:')
            b.write(str(id(self._value)))

    def __repr__(self):
        return '<Name(' + repr(self._value) + ')>'

    def rhs(self):
        """See :meth:`LhsExpression.rhs()`."""
        if self._proper:
            if self._value.is_state():
                return self._value.initial_value()
            elif self._value.lhs() == self:
                return self._value.rhs()
        return None

    def _tree_str(self, b, n):
        b.write(' ' * n + str(self._value) + '\n')

    def _validate(self, trail):
        super()._validate(trail)
        # Check value: String is allowed at construction for debugging, but
        # not here!
        if not self._proper:
            raise IntegrityError(
                'Name value "' + repr(self._value) + '" is not an instance of'
                ' class myokit.Variable', self._token)

    def var(self):
        """See :meth:`LhsExpression.var()`."""
        return self._value


class Derivative(LhsExpression):
    """
    Represents a reference to the time-derivative of a variable.

    *Extends:* :class:`LhsExpression`
    """
    _rbp = FUNCTION_CALL
    _nargs = [1]    # Allows parsing as a function
    __hash__ = LhsExpression.__hash__

    def __init__(self, op):
        super().__init__((op,))
        if not isinstance(op, Name):
            raise IntegrityError(
                'The dot() operator can only be used on variables.',
                self._token)
        self._op = op
        self._proper = self._op._proper
        self._references = set([self])

    def bracket(self, op):
        """See :meth:`Expression.bracket()`."""
        if op != self._op:
            raise ValueError('Given operand is not used in this expression.')
        return False

    def clone(self, subst=None, expand=False, retain=None):
        """See :meth:`Expression.clone()`."""
        if subst and self in subst:
            return subst[self]
        return Derivative(self._op.clone(subst, expand, retain))

    def _code(self, b, c):
        b.write('dot(')
        self._op._code(b, c)
        b.write(')')

    def _diff(self, lhs, idstates):
        # Value isn't a variable? Then always return an object
        if not self._proper:
            return PartialDerivative(self, lhs)

        # Get rhs
        rhs = self._op._value.rhs()

        # Not treating states as independent: then return object if any of the
        # RHS's references are to intermediary or state variables.
        # (Note that the fact that this is a dot(state) doesn't mean it depends
        # on that state: We should instead inspect the RHS.)
        if not idstates:
            for ref in rhs._references:
                var = ref.var()
                if ref == lhs or (not ref._proper) or var.is_state():
                    return PartialDerivative(self, lhs)
                if var.is_intermediary():
                    # Possible state dependency: check!
                    if ref.depends_on_state(deep=True):
                        return PartialDerivative(self, lhs)

        # Check for dependencies in RHS
        if rhs.depends_on(lhs, deep=True):
            return PartialDerivative(self, lhs)
        return None

    def _eval_unit(self, mode):
        # Get numerator (never None in strict mode)
        unit1 = self._op._eval_unit(mode)

        # Get denomenator
        unit2 = \
            myokit.units.dimensionless if mode == myokit.UNIT_STRICT else None
        if self._proper:
            model = self._op._value.model()
            if model is not None:
                unit2 = model.time_unit(mode)

        # Handle as division
        if unit2 is None:
            return unit1    # Can be None in tolerant mode!
        elif unit1 is None:
            return 1 / unit2
        return unit1 / unit2

    def is_derivative(self, var=None):
        """See :meth:`Expression.is_derivative()`."""
        return (var is None) or (var == self._op._value)

    def _polishb(self, b):
        b.write('dot ')
        self._op._polishb(b)

    def __repr__(self):
        return '<Derivative(' + repr(self._op) + ')>'

    def rhs(self):
        """See :meth:`LhsExpression.rhs()`."""
        if self._proper:
            return self._op._value.rhs()
        return None

    def _tree_str(self, b, n):
        b.write(' ' * n + 'dot(' + str(self._op._value) + ')\n')

    def var(self):
        """See :meth:`LhsExpression.var()`."""
        return self._op._value

    def _validate(self, trail):
        super()._validate(trail)
        # Check that value is a variable has already been performed by name
        # Check if value is the name of a state variable
        if not self._op._value.is_state():
            raise IntegrityError(
                'Derivatives can only be defined for state variables.',
                self._token)


class PartialDerivative(LhsExpression):
    """
    Represents a reference to the partial derivative of one variable with
    respect to another.

    This class is used when writing out derivatives of equations, but may _not_
    appear in right-hand-side expressions for model variables!

    *Extends:* :class:`LhsExpression`
    """
    _rbp = FUNCTION_CALL
    _nargs = [2]    # Allows parsing as a function
    __hash__ = LhsExpression.__hash__

    def __init__(self, var1, var2):
        if not isinstance(var1, (Name, Derivative)):
            raise IntegrityError(
                'The first argument to a partial derivative must be a'
                ' variable name or a dot() expression.')
        if not isinstance(var2, (Name, InitialValue)):
            raise IntegrityError(
                'The second argument to a partial derivative must be a'
                ' variable name or an initial value.')
        super().__init__((var1, var2))

        self._var1 = var1
        self._var2 = var2
        self._references = set([self])
        self._has_partials = True

    def bracket(self, op=None):
        """See :meth:`Expression.bracket()`."""
        if op not in self._operands:
            raise ValueError('Given operand is not used in this expression.')
        return False

    def clone(self, subst=None, expand=False, retain=None):
        """See :meth:`Expression.clone()`."""
        if subst and self in subst:
            return subst[self]
        return PartialDerivative(
            self._var1.clone(subst, expand, retain),
            self._var2.clone(subst, expand, retain),
        )

    def _code(self, b, c):
        b.write('diff(')
        self._var1._code(b, c)
        b.write(', ')
        self._var2._code(b, c)
        b.write(')')

    def dependent_expression(self):
        """
        Returns the expression that a derivative is taken of, i.e. "y" in
        "dy/dx".
        """
        return self._var1

    def _diff(self, lhs, idstates):
        raise NotImplementedError(
            'Partial derivatives of partial derivatives are not supported.')

    def _eval_unit(self, mode):
        unit1 = self._var1._eval_unit(mode)
        unit2 = self._var2._eval_unit(mode)
        if unit2 is None:
            return unit1    # Can be None in tolerant mode!
        elif unit1 is None:
            return 1 / unit2
        return unit1 / unit2

    def independent_expression(self):
        """
        Returns the expression that a derivative is taken with respect to, i.e.
        ``x`` in ``dy/dx``.
        """
        return self._var2

    def _polishb(self, b):
        b.write('diff ')
        self._var1._polishb(b)
        self._var2._polishb(b)

    def __repr__(self):
        return ('<PartialDerivative(' + repr(self._var1) + ', '
                + repr(self._var2) + ')>')

    def rhs(self):
        """
        See :meth:`LhsExpression.rhs()`.

        The RHS returned in this case will be ``None``, as there is no RHS
        associated with partial derivatives in the model.
        """
        return None

    def _tree_str(self, b, n):
        b.write(' ' * n + 'partial\n')
        self._var1._tree_str(b, n + self._treeDent)
        self._var2._tree_str(b, n + self._treeDent)

    def var(self):
        """
        See :meth:`LhsExpression.var()`.

        As with time-derivatives, this returns the variable of which a
        derivative is taken (i.e. the dependent variable "y" in "dy/dx").
        """
        return self._var1.var()


class InitialValue(LhsExpression):
    """
    Represents a reference to the initial value of a state variable.

    This class is used when writing out derivatives of equations, but may _not_
    appear in right-hand-side expressions for model variables!

    *Extends:* :class:`LhsExpression`
    """
    _rbp = FUNCTION_CALL
    _nargs = [1]    # Allows parsing as a function
    __hash__ = LhsExpression.__hash__

    def __init__(self, var):
        super().__init__((var, ))
        if not isinstance(var, Name):
            raise IntegrityError(
                'The first argument to an initial value must be a variable'
                ' name.', self._token)

        self._var = var
        self._references = set([self])
        self._has_initials = True

    def bracket(self, op=None):
        """See :meth:`Expression.bracket()`."""
        if op not in self._operands:
            raise ValueError('Given operand is not used in this expression.')
        return False

    def clone(self, subst=None, expand=False, retain=None):
        """See :meth:`Expression.clone()`."""
        if subst and self in subst:
            return subst[self]
        return InitialValue(self._var.clone(subst, expand, retain))

    def _code(self, b, c):
        b.write('init(')
        self._var._code(b, c)
        b.write(')')

    def _diff(self, lhs, idstates):
        raise NotImplementedError(
            'Partial derivatives of initial values are not supported.')

    def _eval_unit(self, mode):
        return self._var._eval_unit(mode)

    def _polishb(self, b):
        b.write('init ')
        self._var._polishb(b)

    def __repr__(self):
        return '<InitialValue(' + repr(self._var) + ')>'

    def rhs(self):
        """
        See :meth:`LhsExpression.rhs()`.

        The RHS returned in this case will be ``None``, as there is no RHS
        associated with initial values in the model.
        """
        # Note: This _could_ return a Number(init, var unit) instead...
        return None

    def _tree_str(self, b, n):
        b.write(' ' * n + 'init(' + str(self._var._value) + ')\n')

    def var(self):
        """See :meth:`LhsExpression.var()`."""
        return self._var._value

    def _validate(self, trail):
        super()._validate(trail)
        # Check if value is the name of a state variable
        var = self._var._value
        if not (isinstance(var, myokit.Variable) and var.is_state()):
            raise IntegrityError(
                'Initial values can only be defined for state variables.',
                self._token)


class PrefixExpression(Expression):
    """
    Base class for prefix expressions: expressions with a single operand.

    *Abstract class, extends:* :class:`Expression`
    """
    _rbp = PREFIX
    _rep = None

    def __init__(self, op):
        super().__init__((op,))
        self._op = op

    def bracket(self, op):
        """See :meth:`Expression.bracket()`."""
        if op != self._op:
            raise ValueError('Given operand is not used in this expression.')
        return (self._op._rbp > LITERAL) and (self._op._rbp < self._rbp)

    def clone(self, subst=None, expand=False, retain=None):
        """See :meth:`Expression.clone()`."""
        if subst and self in subst:
            return subst[self]
        return type(self)(self._op.clone(subst, expand, retain))

    def _code(self, b, c):
        b.write(self._rep)
        brackets = self._op._rbp > LITERAL and self._op._rbp < self._rbp
        if brackets:
            b.write('(')
        self._op._code(b, c)
        if brackets:
            b.write(')')

    def _eval_unit(self, mode):
        return self._op._eval_unit(mode)

    def _tree_str(self, b, n):
        b.write(' ' * n + self._rep + '\n')
        self._op._tree_str(b, n + self._treeDent)


class PrefixPlus(PrefixExpression):
    """
    Prefixed plus. Indicates a positive number ``+op``.

    >>> from myokit import *
    >>> x = PrefixPlus(Number(10))
    >>> print(x.eval())
    10.0

    *Extends:* :class:`PrefixExpression`
    """
    _rep = '+'

    def _diff(self, lhs, idstates):
        return self._op._diff(lhs, idstates)

    def _eval(self, subst, precision):
        try:
            return self._op._eval(subst, precision)
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _polishb(self, b):
        self._op._polishb(b)


class PrefixMinus(PrefixExpression):
    """
    Prefixed minus. Indicates a negative number ``-op``.

    >>> from myokit import *
    >>> x = PrefixMinus(Number(10))
    >>> print(x.eval())
    -10.0

    *Extends:* :class:`PrefixExpression`
    """
    _rep = '-'

    def _diff(self, lhs, idstates):
        op = self._op._diff(lhs, idstates)
        return None if op is None else PrefixMinus(op)

    def _eval(self, subst, precision):
        try:
            return -self._op._eval(subst, precision)
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _polishb(self, b):
        b.write('~ ')
        self._op._polishb(b)


class InfixExpression(Expression):
    """
    Base class for for infix expressions: ``<left> operator <right>``.

    The order of the operands may matter, so that ``<left> operator <right>``
    is not always equal to ``<right> operator <left>``.

    *Abstract class, extends:* :class:`Expression`
    """
    _rep = None      # Operator representation (+, *, et)
    _spaces_round_operator = True

    def __init__(self, left, right):
        super().__init__((left, right))
        self._op1 = left
        self._op2 = right

    def bracket(self, op):
        """See :meth:`Expression.bracket()`."""
        if op == self._op1:
            return (op._rbp > LITERAL and (op._rbp < self._rbp))
        elif op == self._op2:
            return (op._rbp > LITERAL and (op._rbp <= self._rbp))
        raise ValueError('Given operand is not used in this expression.')

    def clone(self, subst=None, expand=False, retain=None):
        """See :meth:`Expression.clone()`."""
        if subst and self in subst:
            return subst[self]
        return type(self)(
            self._op1.clone(subst, expand, retain),
            self._op2.clone(subst, expand, retain))

    def _code(self, b, c):
        # Test bracket locally, avoid function call
        if self._op1._rbp > LITERAL and self._op1._rbp < self._rbp:
            b.write('(')
            self._op1._code(b, c)
            b.write(')')
        else:
            self._op1._code(b, c)
        if self._spaces_round_operator:
            b.write(' ')
        b.write(self._rep)
        if self._spaces_round_operator:
            b.write(' ')
        if self._op2._rbp > LITERAL and self._op2._rbp <= self._rbp:
            b.write('(')
            self._op2._code(b, c)
            b.write(')')
        else:
            self._op2._code(b, c)

    def _polishb(self, b):
        b.write(self._rep)
        b.write(' ')
        self._op1._polishb(b)
        b.write(' ')
        self._op2._polishb(b)

    def _tree_str(self, b, n):
        b.write(' ' * n + self._rep + '\n')
        self._op1._tree_str(b, n + self._treeDent)
        self._op2._tree_str(b, n + self._treeDent)


class Plus(InfixExpression):
    """
    Represents the addition of two operands: ``left + right``.

    >>> from myokit import *
    >>> x = parse_expression('5 + 2')
    >>> print(x.eval())
    7.0

    *Extends:* :class:`InfixExpression`
    """
    _rbp = SUM
    _rep = '+'
    _description = 'Addition'

    def _diff(self, lhs, idstates):
        op1 = self._op1._diff(lhs, idstates)
        op2 = self._op2._diff(lhs, idstates)
        if op1 is None:
            return op2  # Could be None
        if op2 is None:
            return op1  # Definitely not None
        return Plus(op1, op2)

    def _eval(self, subst, precision):
        try:
            return (
                self._op1._eval(subst, precision)
                + self._op2._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _eval_unit(self, mode):
        unit1 = self._op1._eval_unit(mode)
        unit2 = self._op2._eval_unit(mode)

        if unit1 == unit2:
            return unit1
        if unit1 is None:
            return unit2
        if unit2 is None:
            return unit1

        raise EvalUnitError(
            self, 'Addition requires equal units, got '
            + unit1.clarify() + ' and ' + unit2.clarify() + '.')


class Minus(InfixExpression):
    """
    Represents subtraction: ``left - right``.

    >>> from myokit import *
    >>> x = parse_expression('5 - 2')
    >>> print(x.eval())
    3.0

    *Extends:* :class:`InfixExpression`
    """
    _rbp = SUM
    _rep = '-'
    _description = 'Subtraction'

    def _diff(self, lhs, idstates):
        op1 = self._op1._diff(lhs, idstates)
        op2 = self._op2._diff(lhs, idstates)
        if op2 is None:
            return op1  # Could be None
        if op1 is None:
            # Op2 is not None, so need to return -op2
            return PrefixMinus(op2)
        return Minus(op1, op2)

    def _eval(self, subst, precision):
        try:
            return (
                self._op1._eval(subst, precision)
                - self._op2._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _eval_unit(self, mode):
        unit1 = self._op1._eval_unit(mode)
        unit2 = self._op2._eval_unit(mode)

        if unit1 == unit2:
            return unit1
        if unit1 is None:
            return unit2
        if unit2 is None:
            return unit1

        raise EvalUnitError(
            self, 'Subtraction requires equal units, got '
            + unit1.clarify() + ' and ' + unit2.clarify() + '.')


class Multiply(InfixExpression):
    """
    Represents multiplication: ``left * right``.

    >>> from myokit import *
    >>> x = parse_expression('5 * 2')
    >>> print(x.eval())
    10.0

    *Extends:* :class:`InfixExpression`
    """
    _rbp = PRODUCT
    _rep = '*'

    def _diff(self, lhs, idstates):
        op1 = self._op1._diff(lhs, idstates)
        op2 = self._op2._diff(lhs, idstates)
        if op1 is None and op2 is None:
            return None
        elif op2 is None:
            return Multiply(op1, self._op2)     # f' g
        elif op1 is None:
            return Multiply(self._op1, op2)     # f g'
        return Plus(Multiply(op1, self._op2), Multiply(self._op1, op2))

    def _eval(self, subst, precision):
        try:
            return (
                self._op1._eval(subst, precision)
                * self._op2._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _eval_unit(self, mode):
        unit1 = self._op1._eval_unit(mode)
        unit2 = self._op2._eval_unit(mode)

        if unit1 is None:
            return unit2
        if unit2 is None:
            return unit1

        return unit1 * unit2


class Divide(InfixExpression):
    """
    Represents division: ``left / right``.

    >>> from myokit import *
    >>> x = parse_expression('5 / 2')
    >>> print(x.eval())
    2.5

    *Extends:* :class:`InfixExpression`
    """
    _rbp = PRODUCT
    _rep = '/'

    def _diff(self, lhs, idstates):
        op1 = self._op1._diff(lhs, idstates)
        op2 = self._op2._diff(lhs, idstates)

        if op1 is None and op2 is None:
            return None
        elif op2 is None:
            # g f' / g^2 = f' / g
            return Divide(op1, self._op2)
        elif op1 is None:
            # -(f g') / g^2
            return Divide(
                Multiply(PrefixMinus(self._op1), op2),
                Power(self._op2, Number(2))
            )

        # (f' g - f g') / g^2
        return Divide(
            Minus(Multiply(op1, self._op2), Multiply(self._op1, op2)),
            Power(self._op2, Number(2))
        )

    def _eval(self, subst, precision):
        try:
            b = self._op2._eval(subst, precision)
            if b == 0:
                raise ZeroDivisionError()
            return self._op1._eval(subst, precision) / b
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _eval_unit(self, mode):
        unit1 = self._op1._eval_unit(mode)
        unit2 = self._op2._eval_unit(mode)

        if unit2 is None:
            return unit1    # None propagation in tolerant mode
        elif unit1 is None:
            return 1 / unit2

        return unit1 / unit2


class Quotient(InfixExpression):
    """
    Represents the quotient of a division ``left // right``, also known as
    integer division.

    >>> import myokit
    >>> x = myokit.parse_expression('7 // 3')
    >>> print(x.eval())
    2.0

    Note that, for negative numbers Myokit follows the convention of rounding
    towards negative infinity, rather than towards zero. Thus:

    >>> print(myokit.parse_expression('-7 // 3').eval())
    -3.0

    Similarly:

    >>> print(myokit.parse_expression('5 // -3').eval())
    -2.0

    Note that this differs from how integer division is implemented in C, which
    _truncates_ (round towards zero), but similar to how ``floor()`` is
    implemented in C, which rounds towards negative infinity.

    See: https://python-history.blogspot.co.uk/2010/08/
    And: https://en.wikipedia.org/wiki/Euclidean_division

    *Extends:* :class:`InfixExpression`
    """
    _rbp = PRODUCT
    _rep = '//'

    def _diff(self, lhs, idstates):
        # The result of a // b is always flat, with discontinuous jumps
        # whenever a = k*b (for an integer k). As a result, the derivatives
        # d/da(a//b) and d/db(a//b) are either zero (most of the time) or
        # undefined (at the jumps). Notably, outside of the jump points, the
        # line a//b is flat, so does not depend on a, b, a', or b'!
        #
        # Alternatively, a // b = floor(a / b).
        #
        # Here we ignore the discontinuities in favor of a left or right
        # derivative, and simply return zero for all points.
        return None

    def _eval(self, subst, precision):
        try:
            return (
                self._op1._eval(subst, precision)
                // self._op2._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _eval_unit(self, mode):
        unit1 = self._op1._eval_unit(mode)
        unit2 = self._op2._eval_unit(mode)

        if unit2 is None:
            return unit1    # None propagation in tolerant mode
        elif unit1 is None:
            return 1 / unit2

        return unit1 / unit2


class Remainder(InfixExpression):
    """
    Represents the remainder of a division (the "modulo"), expressed in ``mmt``
    syntax as ``left % right``.

    >>> import myokit
    >>> x = myokit.parse_expression('7 % 3')
    >>> print(x.eval())
    1.0

    Note that, for negative numbers Myokit follows the convention of Python
    that the quotient is rounded to negative infinity. Thus:

    >>> print(myokit.parse_expression('-7 // 3').eval())
    -3.0

    and therefore:

    >>> print(myokit.parse_expression('-7 % 3').eval())
    2.0

    Similarly:

    >>> print(myokit.parse_expression('5 // -3').eval())
    -2.0

    so that:

    >>> print(myokit.parse_expression('5 % -3').eval())
    -1.0

    See: https://en.wikipedia.org/wiki/Modulo_operation

    *Extends:* :class:`InfixExpression`
    """
    _rbp = PRODUCT
    _rep = '%'

    def _diff(self, lhs, idstates):
        # Since
        #   a(x) % b(x) is defined as a(x) - b(x) * floor(a(x) / b(x))
        # its derivative is
        #   a' - b' floor(a/b) - b * d/dx floor(a/b).
        # Using d/dx floor(a/b) = 0 (ignoring discontinuities, see Floor), that
        # simplifies to
        #   a' - b' floor(a/b)

        op1 = self._op1._diff(lhs, idstates)
        op2 = self._op2._diff(lhs, idstates)

        if op1 is None and op2 is None:
            return None
        elif op1 is None:
            # -b' floor(a/b)
            return Multiply(
                PrefixMinus(op2), Floor(Divide(self._op1, self._op2)))
        elif op2 is None:
            # a'
            return op1

        # a' - b' floor(a/b)
        return Minus(op1, Multiply(op2, Floor(Divide(self._op1, self._op2))))

    def _eval(self, subst, precision):
        try:
            return (
                self._op1._eval(subst, precision)
                % self._op2._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _eval_unit(self, mode):
        # 14 pizzas / 5 kids = 2 pizzas / kid + 4 pizzas
        unit1 = self._op1._eval_unit(mode)
        self._op2._eval_unit(mode)  # Also check op2!
        return unit1


class Power(InfixExpression):
    """
    Represents exponentiation: ``left ^ right``.

    >>> import myokit
    >>> x = myokit.parse_expression('5 ^ 2')
    >>> print(x.eval())
    25.0

    *Extends:* :class:`InfixExpression`
    """
    _rbp = POWER
    _rep = '^'
    _spaces_round_operator = False

    def _diff(self, lhs, idstates):
        # The general rule is derived using a(x)^b(x) = e^(ln(a(x)) * b(x)).
        # This only works when ln(a(x)) is defined, so when a(x) >= 0. But this
        # is OK, as a(x)^b(x) is only defined for fractional b(x) if a(x) >= 0!
        #
        # Applying the chain and product rules:
        #  d/dx e^(ln(a) * b)
        #   = e^(ln(a) * b) * d/dx (ln(a) * b)
        #   = a^b * (ln(a) * b' + b / a * a')
        #
        # If b' is 0 this reduces to a^b * b/a * a' = b * a^(b-1) * a'
        # If a' is 0 this reduces to a^b * b' / ln(a) (for a >= 0)
        #
        # If b depends on the lhs, to have a derivative it _must_ be
        # fractional, and so a(x) _must_ be positive. We don't need to check
        # this; it will just fail along with normal evaluation.
        # If b does not depend on the lhs, we use the reduced form i.e. the
        # general power rule.
        #
        op1 = self._op1._diff(lhs, idstates)
        op2 = self._op2._diff(lhs, idstates)

        if op1 is None and op2 is None:
            return None

        if op2 is None:
            # Tweaks: x^number --> reduce number by 1; and x^2 becomes x
            if isinstance(self._op2, Number):
                if self._op2.value() == 2:
                    new_power = self._op1
                elif self._op2.value() == 1:
                    return op1
                else:
                    new_power = Number(self._op2.value() - 1, self._op2.unit())
                    new_power = Power(self._op1, new_power)
            else:
                new_power = Power(self._op1, Minus(self._op2, Number(1)))
            return Multiply(Multiply(self._op2, new_power), op1)

        if op1 is None:
            return Divide(Multiply(self, op2), Log(self._op1))

        return Multiply(self, Plus(
            Multiply(Log(self._op1), op2),
            Multiply(Divide(self._op2, self._op1), op1)
        ))

    def _eval(self, subst, precision):
        try:
            return (
                self._op1._eval(subst, precision)
                ** self._op2._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _eval_unit(self, mode):
        unit1 = self._op1._eval_unit(mode)
        unit2 = self._op2._eval_unit(mode)

        # In strict mode, check 2nd unit is dimensionless
        if unit2 != myokit.units.dimensionless and mode == myokit.UNIT_STRICT:
            raise EvalUnitError(
                self,
                'Exponent in Power must be dimensionless, got '
                + unit2.clarify() + '.')

        if unit1 is None:
            return None

        return unit1 ** self._op2.eval()


class Function(Expression):
    """
    Base class for built-in functions.

    Functions have a ``name``, which must be set to a human readable function
    name (usually just the function's ``.mmt`` equivalent) and a list of
    integers called ``_nargs``. Each entry in ``_nargs`` specifies a number of
    arguments for which the function is implemented. For example :class:`Sin`
    has ``_nargs=[1]`` while :class:`Log` has ``_nargs=[1,2]``, showing that
    ``Log`` can be called with either ``1`` or ``2`` arguments.

    If errors occur when creating a function, an IntegrityError may be thrown.

    *Abstract class, extends:* :class:`Expression`
    """
    _nargs = [1]
    _fname = None
    _rbp = FUNCTION_CALL

    def __init__(self, *ops):
        super().__init__(ops)
        if self._nargs is not None:
            if not len(ops) in self._nargs:
                raise IntegrityError(
                    'Function (' + str(self._fname) + ') created with wrong'
                    ' number of arguments (' + str(len(ops)) + ', expecting '
                    + ' or '.join([str(x) for x in self._nargs]) + ').',
                    self._token)

    def bracket(self, op=None):
        """See :meth:`Expression.bracket()`."""
        if op not in self._operands:
            raise ValueError('Given operand is not used in this expression.')
        return False

    def clone(self, subst=None, expand=False, retain=None):
        """See :meth:`Expression.clone()`."""
        if subst and self in subst:
            return subst[self]
        return type(self)(
            *[x.clone(subst, expand, retain) for x in self._operands])

    def _code(self, b, c):
        b.write(self._fname)
        b.write('(')
        if len(self._operands) > 0:
            self._operands[0]._code(b, c)
            for i in range(1, len(self._operands)):
                b.write(', ')
                self._operands[i]._code(b, c)
        b.write(')')

    def _polishb(self, b):
        # Function name | Number of operands | operands
        # This is sufficient for what we're doing here :)
        b.write(self._fname)
        b.write(' ')
        b.write(str(len(self._operands)))
        for op in self._operands:
            b.write(' ')
            op._polishb(b)

    def _tree_str(self, b, n):
        b.write(' ' * n + self._fname + '\n')
        for op in self._operands:
            op._tree_str(b, n + self._treeDent)


class UnaryDimensionlessFunction(Function):
    """
    Function with a single operand that has dimensionless input and output.

    *Abstract class, extends:* :class:`Function`
    """
    def _eval_unit(self, mode):
        unit = self._operands[0]._eval_unit(mode)

        # Propagate None in tolerant mode
        if unit is None:
            return None

        # Check unit in strict mode
        if mode == myokit.UNIT_STRICT and unit != myokit.units.dimensionless:
            raise EvalUnitError(
                self, 'Function ' + self._fname + '() requires a'
                ' dimensionless operand, got ' + unit.clarify() + '.')

        # Unary dimensionless functions are always dimensionless
        return myokit.units.dimensionless


class Sqrt(Function):
    """
    Represents the square root ``sqrt(x)``.

    >>> import myokit
    >>> x = myokit.parse_expression('sqrt(25)')
    >>> print(x.eval())
    5.0

    *Extends:* :class:`Function`
    """
    _fname = 'sqrt'

    def _diff(self, lhs, idstates):
        op = self._operands[0]
        dop = op._diff(lhs, idstates)
        if dop is None:
            return None
        return Divide(dop, Multiply(Number(2), self))

    def _eval(self, subst, precision):
        try:
            return numpy.sqrt(self._operands[0]._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _eval_unit(self, mode):
        unit = self._operands[0]._eval_unit(mode)
        if unit is None:
            return None
        return unit ** 0.5


class Sin(UnaryDimensionlessFunction):
    """
    Represents the sine function ``sin(x)``.

    >>> from myokit import *
    >>> x = parse_expression('sin(0)')
    >>> print(round(x.eval(), 1))
    0.0
    >>> x = Sin(Number(3.1415 / 2.0))
    >>> print(round(x.eval(), 1))
    1.0

    *Extends:* :class:`UnaryDimensionlessFunction`
    """
    _fname = 'sin'

    def _diff(self, lhs, idstates):
        op = self._operands[0]
        dop = op._diff(lhs, idstates)
        if dop is None:
            return None
        return Multiply(Cos(op), dop)

    def _eval(self, subst, precision):
        try:
            return numpy.sin(self._operands[0]._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)


class Cos(UnaryDimensionlessFunction):
    """
    Represents the cosine function ``cos(x)``.

    >>> from myokit import *
    >>> x = Cos(Number(0))
    >>> print(round(x.eval(), 1))
    1.0
    >>> x = Cos(Number(3.1415 / 2.0))
    >>> print(round(x.eval(), 1))
    0.0

    *Extends:* :class:`UnaryDimensionlessFunction`
    """
    _fname = 'cos'

    def _diff(self, lhs, idstates):
        op = self._operands[0]
        dop = op._diff(lhs, idstates)
        if dop is None:
            return None
        return Multiply(PrefixMinus(Sin(op)), dop)

    def _eval(self, subst, precision):
        try:
            return numpy.cos(self._operands[0]._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)


class Tan(UnaryDimensionlessFunction):
    """
    Represents the tangent function ``tan(x)``.

    >>> from myokit import *
    >>> x = Tan(Number(3.1415 / 4.0))
    >>> print(round(x.eval(), 1))
    1.0

    *Extends:* :class:`UnaryDimensionlessFunction`
    """
    _fname = 'tan'

    def _diff(self, lhs, idstates):
        op = self._operands[0]
        dop = op._diff(lhs, idstates)
        if dop is None:
            return None
        return Divide(dop, Power(Cos(op), Number(2)))

    def _eval(self, subst, precision):
        try:
            return numpy.tan(self._operands[0]._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)


class ASin(UnaryDimensionlessFunction):
    """
    Represents the inverse sine function ``asin(x)``.

    >>> from myokit import *
    >>> x = ASin(Sin(Number(1)))
    >>> print(round(x.eval(), 1))
    1.0

    *Extends:* :class:`UnaryDimensionlessFunction`
    """
    _fname = 'asin'

    def _diff(self, lhs, idstates):
        op = self._operands[0]
        dop = op._diff(lhs, idstates)
        if dop is None:
            return None
        return Divide(dop, Sqrt(Minus(Number(1), Power(op, Number(2)))))

    def _eval(self, subst, precision):
        try:
            return numpy.arcsin(self._operands[0]._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)


class ACos(UnaryDimensionlessFunction):
    """
    Represents the inverse cosine ``acos(x)``.

    >>> from myokit import *
    >>> x = ACos(Cos(Number(3)))
    >>> print(round(x.eval(), 1))
    3.0

    *Extends:* :class:`UnaryDimensionlessFunction`
    """
    _fname = 'acos'

    def _diff(self, lhs, idstates):
        op = self._operands[0]
        dop = op._diff(lhs, idstates)
        if dop is None:
            return None
        return Divide(
            PrefixMinus(dop),
            Sqrt(Minus(Number(1), Power(op, Number(2))))
        )

    def _eval(self, subst, precision):
        try:
            return numpy.arccos(self._operands[0]._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)


class ATan(UnaryDimensionlessFunction):
    """
    Represents the inverse tangent function ``atan(x)``.

    >>> from myokit import *
    >>> x = ATan(Tan(Number(1)))
    >>> print(round(x.eval(), 1))
    1.0

    If two arguments are given they are interpreted as the coordinates of a
    point (x, y) and the function returns this point's angle with the
    (positive) x-axis. In this case, the returned value will be in the range
    (-pi, pi].

    *Extends:* :class:`UnaryDimensionlessFunction`
    """
    _fname = 'atan'

    def _diff(self, lhs, idstates):
        op = self._operands[0]
        dop = op._diff(lhs, idstates)
        if dop is None:
            return None
        return Divide(dop, Plus(Number(1), Power(op, Number(2))))

    def _eval(self, subst, precision):
        try:
            return numpy.arctan(self._operands[0]._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)


class Exp(UnaryDimensionlessFunction):
    """
    Represents a power of *e*. Written ``exp(x)`` in ``.mmt`` syntax.

    >>> from myokit import *
    >>> x = Exp(Number(1))
    >>> print(round(x.eval(), 4))
    2.7183

    *Extends:* :class:`UnaryDimensionlessFunction`
    """
    _fname = 'exp'

    def _diff(self, lhs, idstates):
        op = self._operands[0]
        dop = op._diff(lhs, idstates)
        if dop is None:
            return None
        return Multiply(self, dop)

    def _eval(self, subst, precision):
        try:
            return numpy.exp(self._operands[0]._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)


class Log(Function):
    """
    With one argument ``log(x)`` represents the natural logarithm.
    With two arguments ``log(x, k)`` is taken to be the base ``k`` logarithm of
    ``x``.

    >>> from myokit import *
    >>> x = Log(Number(10))
    >>> print(round(x.eval(), 4))
    2.3026
    >>> x = Log(Exp(Number(10)))
    >>> print(round(x.eval(), 1))
    10.0
    >>> x = Log(Number(256), Number(2))
    >>> print(round(x.eval(), 1))
    8.0

    *Extends:* :class:`Function`
    """
    _fname = 'log'
    _nargs = [1, 2]

    def _diff(self, lhs, idstates):

        if len(self._operands) == 1:
            # One operand: natural logarithm: a' / a
            op = self._operands[0]
            dop = op._diff(lhs, idstates)
            if dop is None:
                return None
            return Divide(dop, op)

        else:
            # Two operands: log_a(b) = Log(b, a)
            op1 = self._operands[0]     # b
            op2 = self._operands[1]     # a
            dop1 = op1._diff(lhs, idstates)     # b'
            dop2 = op2._diff(lhs, idstates)     # a'
            if dop1 is None and dop2 is None:
                return None
            elif dop2 is None:
                # a' = 0 --> b' / (b * ln(a))
                return Divide(dop1, Multiply(op1, Log(op2)))

            elif dop1 is None:
                # b' = 0 --> -a' ln(b) / (a ln(a)^2)
                return Divide(
                    Multiply(PrefixMinus(dop2), Log(op1)),
                    Multiply(op2, Power(Log(op2), Number(2))),
                )

            # Full form:
            #   b' / (b * ln(a)) - (a' ln(b)) / (a ln(a)^2)
            return Minus(
                Divide(dop1, Multiply(op1, Log(op2))),
                Divide(
                    Multiply(dop2, Log(op1)),
                    Multiply(op2, Power(Log(op2), Number(2))),
                )
            )

    def _eval(self, subst, precision):
        try:
            if len(self._operands) == 1:
                return numpy.log(self._operands[0]._eval(subst, precision))
            return (
                numpy.log(self._operands[0]._eval(subst, precision)) /
                numpy.log(self._operands[1]._eval(subst, precision)))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _eval_unit(self, mode):

        # Check contents of single operand
        if len(self._operands) == 1:
            unit = self._operands[0]._eval_unit(mode)

            # Propagate None in tolerant mode
            if unit is None:
                return None

            # Check units in strict mode
            if mode == myokit.UNIT_STRICT:
                if unit != myokit.units.dimensionless:
                    raise EvalUnitError(
                        self,
                        'Log() requires a dimensionless operand, got '
                        + unit.clarify() + '.')

        # Two operands
        else:
            unit1 = self._operands[0]._eval_unit(mode)
            unit2 = self._operands[1]._eval_unit(mode)

            # Propagate None in tolerant mode
            if unit1 is None and unit2 is None:
                return None

            # Check units in strict mode
            if mode == myokit.UNIT_STRICT:
                if not unit1 == unit2 == myokit.units.dimensionless:
                    raise EvalUnitError(
                        self,
                        'Log() requires dimensionless operands, got '
                        + unit1.clarify() + ' and ' + unit2.clarify() + '.')

        return myokit.units.dimensionless


class Log10(UnaryDimensionlessFunction):
    """
    Represents the base-10 logarithm ``log10(x)``.

    >>> from myokit import *
    >>> x = Log10(Number(100))
    >>> print(round(x.eval(), 1))
    2.0

    *Extends:* :class:`UnaryDimensionlessFunction`
    """
    _fname = 'log10'

    def _diff(self, lhs, idstates):
        op = self._operands[0]
        dop = op._diff(lhs, idstates)
        if dop is None:
            return None
        return Divide(dop, Multiply(op, Log(Number(10))))

    def _eval(self, subst, precision):
        try:
            return numpy.log10(self._operands[0]._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)


class Floor(Function):
    """
    Represents a rounding towards minus infinity ``floor(x)``.

    >>> from myokit import *
    >>> x = Floor(Number(5.2))
    >>> print(x.eval())
    5.0
    >>> x = Floor(Number(-5.2))
    >>> print(x.eval())
    -6.0

    *Extends:* :class:`Function`
    """
    _fname = 'floor'

    def _diff(self, lhs, idstates):
        # Floor is stepwise constant, so it's derivative is zero except when
        # the value of its operand is an integer. Here we simplify and just say
        # the derivative is always zero.
        return None

    def _eval(self, subst, precision):
        try:
            return numpy.floor(self._operands[0]._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _eval_unit(self, mode):
        return self._operands[0]._eval_unit(mode)


class Ceil(Function):
    """
    Represents a rounding towards positve infinity ``ceil(x)``.

    >>> from myokit import *
    >>> x = Ceil(Number(5.2))
    >>> print(x.eval())
    6.0
    >>> x = Ceil(Number(-5.2))
    >>> print(x.eval())
    -5.0

    *Extends:* :class:`Function`
    """
    _fname = 'ceil'

    def _diff(self, lhs, idstates):
        # Ceil is stepwise constant, so it's derivative is zero except when
        # the value of its operand is an integer. Here we simplify and just say
        # the derivative is always zero.
        return None

    def _eval(self, subst, precision):
        try:
            return numpy.ceil(self._operands[0]._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _eval_unit(self, mode):
        return self._operands[0]._eval_unit(mode)


class Abs(Function):
    """
    Returns the absolute value of a number ``abs(x)``.

    >>> from myokit import *
    >>> x = parse_expression('abs(5)')
    >>> print(x.eval())
    5.0
    >>> x = parse_expression('abs(-5)')
    >>> print(x.eval())
    5.0

    *Extends:* :class:`Function`
    """
    _fname = 'abs'

    def _diff(self, lhs, idstates):
        # The derivative of abs(f(x)) is f'(x) for x > 0, -f'(x) for x < 0, and
        # undefined if x == 0. Here we simplify and say it's f'(x) if x >= 0.

        # Get derivative of operand
        op = self._operands[0]
        dop = op._diff(lhs, idstates)
        if dop is None:
            return None

        # Get operand unit (_not_ the unit of the returned derivative)
        try:
            unit = op.eval_unit(myokit.UNIT_TOLERANT)
        except myokit.IncompatibleUnitError:
            unit = None

        # Return if
        return If(MoreEqual(op, Number(0, unit)), dop, PrefixMinus(dop))

    def _eval(self, subst, precision):
        try:
            return numpy.abs(self._operands[0]._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _eval_unit(self, mode):
        return self._operands[0]._eval_unit(mode)


class If(Function):
    """
    Allows conditional functions to be defined using an if-then-else structure.

    The first argument to an ``If`` function must be a condition, followed
    directly by an expression to use to calculate the function's return value
    if this condition is ``True``. The third and final argument specifies the
    expression's value if the condition is ``False``.

    A simple example in ``.mmt`` syntax::

        x = if(V < 10, 5 * V + 100, 6 * V)

    *Extends:* :class:`Function`
    """
    _nargs = [3]
    _fname = 'if'

    def __init__(self, i, t, e):
        super().__init__(i, t, e)
        self._i = i     # if
        self._t = t     # then
        self._e = e     # else

    def condition(self):
        """
        Returns this if-function's condition.
        """
        return self._i

    def _diff(self, lhs, idstates):
        op1 = self._t._diff(lhs, idstates)
        op2 = self._e._diff(lhs, idstates)

        # Return None if both None, or if() if both expressions
        if op1 is None and op2 is None:
            return None
        elif op1 is not None and op2 is not None:
            return If(self._i, op1, op2)

        # If only one is None, create a zero with the correct units
        zero = Number(0, self._diff_unit(lhs))
        if op1 is None:
            return If(self._i, zero, op2)
        else:
            return If(self._i, op1, zero)

    def _eval(self, subst, precision):
        if self._i._eval(subst, precision):
            return self._t._eval(subst, precision)
        return self._e._eval(subst, precision)

    def _eval_unit(self, mode):

        # Check the condition and all options
        self._i._eval_unit(mode)
        unit2 = self._t._eval_unit(mode)
        unit3 = self._e._eval_unit(mode)

        # Check if the options have the same unit (or None in tolerant mode)
        if unit2 == unit3:
            return unit2

        # Allow 1 None in tolerant mode
        if unit2 is None:
            return unit3
        elif unit3 is None:
            return unit2

        # Mismatching units
        raise EvalUnitError(
            self, 'Units of `then` and `else` part of an `if`'
            ' must match. Got ' + str(unit2) + ' and ' + str(unit3) + '.')

    def is_conditional(self):
        return True

    def piecewise(self):
        """
        Returns an equivalent ``Piecewise`` object.
        """
        return Piecewise(self._i, self._t, self._e)

    def value(self, which):
        """
        Returns the expression used when this if's condition is ``True`` when
        called with ``which=True``. Otherwise return the expression used when
        this if's condition is ``False``.
        """
        return self._t if which else self._e


class Piecewise(Function):
    """
    Allows piecewise functions to be defined.

    The first argument to a ``Piecewise`` function must be a condition,
    followed directly by an expression to use to calculate the function's
    return value if this condition is true.

    Any number of condition-expression pairs can be added. If multiple
    conditions evaluate as true, only the first one will be used to set the
    return value.

    The final argument should be a default expression to use if none of the
    conditions evaluate to True. This means the ``piecewise()`` function can
    have any odd number of arguments greater than 2.

    A simple example in ``mmt`` syntax::

        x = piecewise(
            V < 10, 5 * V + 100
            V < 20, 6 * V,
            7 * V)

    This will return ``5 * V + 100`` for any value smaller than 10, ``6 * V``
    for any value greater or equal to 10 but smaller than 20, and ``7 * V`` for
    any values greather than or equal to 20.

    *Extends:* :class:`Function`
    """
    _nargs = None
    _fname = 'piecewise'

    def __init__(self, *ops):
        super().__init__(*ops)

        # Check number of arguments
        n = len(self._operands)
        if n % 2 == 0:
            raise IntegrityError(
                'Piecewise function must have odd number of arguments:'
                ' ([condition, value]+, else_value).', self._token)
        if n < 3:
            raise IntegrityError(
                'Piecewise function must have 3 or more arguments.',
                self._token)

        # Check arguments
        m = n // 2
        self._i = [0] * m           # Conditions
        self._e = [0] * (m + 1)     # Expressions
        oper = iter(ops)
        for i in range(0, m):
            self._i[i] = next(oper)
            self._e[i] = next(oper)
        self._e[m] = next(oper)

    def conditions(self):
        """
        Returns an iterator over the conditions used by this Piecewise.
        """
        return iter(self._i)

    def _diff(self, lhs, idstates):

        # Evaluate derivatives of the (m + 1) expressions
        ops = [op._diff(lhs, idstates) for op in self._e]

        # Count the number of Nones
        n_zeroes = sum([1 if op is None else 0 for op in ops])

        # Return None if all None
        if n_zeroes == len(ops):
            return None

        # Create zero to use if any of the expressions are None
        zero = None
        if n_zeroes > 0:
            zero = Number(0, self._diff_unit(lhs))

        # Create and return piecewise
        new_ops = [0] * (2 * len(self._i) + 1)
        for i, _if in enumerate(self._i):
            new_ops[2 * i] = _if
            new_ops[2 * i + 1] = zero if ops[i] is None else ops[i]
        new_ops[-1] = zero if ops[-1] is None else ops[-1]
        return Piecewise(*new_ops)

    def _eval(self, subst, precision):
        for k, cond in enumerate(self._i):
            if cond._eval(subst, precision):
                return self._e[k]._eval(subst, precision)
        return self._e[-1]._eval(subst, precision)

    def _eval_unit(self, mode):

        # Check the conditions and all options
        units = [x._eval_unit(mode) for x in self._i]   # And discard :)
        units = [x._eval_unit(mode) for x in self._e]

        # Check if the options have the same unit
        units = set(units)
        # Nones are allowed in tolerant mode, can't occur in strict mode
        if None in units:
            units.remove(None)
            if len(units) == 0:
                return None
        if len(units) == 1:
            return units.pop()
        raise EvalUnitError(
            self, 'All branches of a piecewise() must have the same unit.')

    def is_conditional(self):
        return True

    def pieces(self):
        """
        Returns an iterator over the pieces in this Piecewise.
        """
        return iter(self._e)


class Condition:
    """
    *Abstract class*

    Interface for conditional expressions that can be evaluated to True or
    False. Doesn't add any methods but simply indicates that this is a
    condition.
    """

    def _diff(self, lhs, idstates):
        raise NotImplementedError(
            'Conditions do not have partial derivatives.')


class PrefixCondition(PrefixExpression, Condition):
    """
    Interface for prefix conditions.

    *Abstract class, extends:* :class:`Condition`, :class:`PrefixExpression`
    """


class Not(PrefixCondition):
    """
    Negates a condition. Written as ``not x``.

    >>> from myokit import *
    >>> x = parse_expression('1 == 1')
    >>> print(x.eval())
    True
    >>> y = Not(x)
    >>> print(y.eval())
    False
    >>> x = parse_expression('(2 == 2) and not (1 > 2)')
    >>> print(x.eval())
    True

    *Extends:* :class:`PrefixCondition`
    """
    _rep = 'not'

    def _code(self, b, c):
        b.write('not ')
        brackets = self._op._rbp > LITERAL and self._op._rbp < self._rbp
        if brackets:
            b.write('(')
        self._op._code(b, c)
        if brackets:
            b.write(')')

    def _eval(self, subst, precision):
        try:
            return not self._op._eval(subst, precision)
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _eval_unit(self, mode):
        unit = self._op._eval_unit(mode)
        if unit not in (None, myokit.units.dimensionless):
            raise EvalUnitError(
                self, 'Operator `not` expects a dimensionless operand.')
        return unit

    def _polishb(self, b):
        b.write('not ')
        self._op._polishb(b)


class InfixCondition(InfixExpression, Condition):
    """
    Base class for infix expressions.

    *Abstract class, extends:* :class:`Condition`, :class:`InfixExpression`
    """
    _rbp = CONDITIONAL


class BinaryComparison(InfixCondition):
    """
    Base class for infix comparisons of two entities.

    *Abstract class, extends:* :class:`InfixCondition`
    """
    def _eval_unit(self, mode):
        unit1 = self._op1._eval_unit(mode)
        unit2 = self._op2._eval_unit(mode)

        # Equal (including both None) is always ok
        if unit1 == unit2:
            return None if unit1 is None else myokit.units.dimensionless

        # In tolerant mode, a single None is OK
        if unit1 is None or unit2 is None:
            return myokit.units.dimensionless

        # Otherwise must match
        raise EvalUnitError(
            self, 'Condition ' + self._rep + ' requires equal units on both'
            ' sides, got ' + str(unit1) + ' and ' + str(unit2) + '.')


class Equal(BinaryComparison):
    """
    Represents an equality check ``x == y``.

    >>> from myokit import *
    >>> print(parse_expression('1 == 0').eval())
    False
    >>> print(parse_expression('1 == 1').eval())
    True

    *Extends:* :class:`InfixCondition`
    """
    _rep = '=='

    def _eval(self, subst, precision):
        try:
            return (
                self._op1._eval(subst, precision)
                == self._op2._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)


class NotEqual(BinaryComparison):
    """
    Represents an inequality check ``x != y``.

    >>> from myokit import *
    >>> print(parse_expression('1 != 0').eval())
    True
    >>> print(parse_expression('1 != 1').eval())
    False

    *Extends:* :class:`InfixCondition`
    """
    _rep = '!='

    def _eval(self, subst, precision):
        try:
            return (
                self._op1._eval(subst, precision)
                != self._op2._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)


class More(BinaryComparison):
    """
    Represents an is-more-than check ``x > y``.

    >>> from myokit import *
    >>> print(parse_expression('5 > 2').eval())
    True

    *Extends:* :class:`InfixCondition`
    """
    _rep = '>'

    def _eval(self, subst, precision):
        try:
            return (
                self._op1._eval(subst, precision)
                > self._op2._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)


class Less(BinaryComparison):
    """
    Represents an is-less-than check ``x < y``.

    >>> from myokit import *
    >>> print(parse_expression('5 < 2').eval())
    False

    *Extends:* :class:`InfixCondition`
    """
    _rep = '<'

    def _eval(self, subst, precision):
        try:
            return (
                self._op1._eval(subst, precision)
                < self._op2._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)


class MoreEqual(BinaryComparison):
    """
    Represents an is-more-than-or-equal check ``x > y``.

    >>> from myokit import *
    >>> print(parse_expression('2 >= 2').eval())
    True

    *Extends:* :class:`InfixCondition`
    """
    _rep = '>='

    def _eval(self, subst, precision):
        try:
            return (
                self._op1._eval(subst, precision)
                >= self._op2._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)


class LessEqual(BinaryComparison):
    """
    Represents an is-less-than-or-equal check ``x <= y``.

    >>> from myokit import *
    >>> print(parse_expression('2 <= 2').eval())
    True

    *Extends:* :class:`InfixCondition`
    """
    _rep = '<='

    def _eval(self, subst, precision):
        try:
            return (
                self._op1._eval(subst, precision)
                <= self._op2._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)


class And(InfixCondition):
    """
    True if two conditions are true: ``x and y``.

    >>> from myokit import *
    >>> print(parse_expression('1 == 1 and 2 == 4').eval())
    False
    >>> print(parse_expression('1 == 1 and 4 == 4').eval())
    True

    *Extends:* :class:`InfixCondition`
    """
    _rbp = CONDITION_AND
    _rep = 'and'

    def _eval(self, subst, precision):
        try:
            return (
                self._op1._eval(subst, precision)
                and self._op2._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _eval_unit(self, mode):
        unit1 = self._op1._eval_unit(mode)
        unit2 = self._op2._eval_unit(mode)

        # Propagate both None in tolerant mode
        if unit1 is None and unit2 is None:
            return None

        # Ideal: both dimensionless
        unit1 = myokit.units.dimensionless if unit1 is None else unit1
        unit2 = myokit.units.dimensionless if unit2 is None else unit2
        if unit1 == unit2 == myokit.units.dimensionless:
            return unit1

        raise EvalUnitError(
            self, 'Operator `and` expects dimensionless operands.')


class Or(InfixCondition):
    """
    True if at least one of two conditions is true: ``x or y``.

    >>> from myokit import *
    >>> print(parse_expression('1 == 1 or 2 == 4').eval())
    True

    *Extends:* :class:`InfixCondition`
    """
    _rbp = CONDITION_AND
    _rep = 'or'

    def _eval(self, subst, precision):
        try:
            return (
                self._op1._eval(subst, precision)
                or self._op2._eval(subst, precision))
        except (ArithmeticError, ValueError) as e:  # pragma: no cover
            raise EvalError(self, subst, e)

    def _eval_unit(self, mode):
        unit1 = self._op1._eval_unit(mode)
        unit2 = self._op2._eval_unit(mode)

        # Propagate both None in tolerant mode
        if unit1 is None and unit2 is None:
            return None

        # Ideal: both dimensionless
        unit1 = myokit.units.dimensionless if unit1 is None else unit1
        unit2 = myokit.units.dimensionless if unit2 is None else unit2
        if unit1 == unit2 == myokit.units.dimensionless:
            return unit1

        raise EvalUnitError(
            self, 'Operator `or` expects dimensionless operands.')


class EvalError(Exception):
    """
    Used internally when an error is encountered during an ``eval()``
    operation. Is replaced by a ``myokit.NumericalError`` which is then sent
    to the caller.

    Attributes:

    ``expr``
        The expression that generated the error
    ``subst``
        Any substitutions given to the eval function
    ``err``
        The exception that triggered the error or an error message

    *Extends:* ``Exception``
    """
    def __init__(self, expr, subst, err):
        self.expr = expr
        self.err = err


class EvalUnitError(Exception):
    """
    Used internally when an error is encountered during an ``eval_unit()``
    operation. Is replaced by a :class:`myokit.IncompatibleUnitError` which is
    then sent to the caller.

    ``expr``
        The expression that generated the error
    ``err``
        The exception that triggered the error or an error message

    *Extends:* ``Exception``
    """
    def __init__(self, expr, err):
        self.expr = expr
        self.err = err
        self.line = self.char = None
        if expr._token is not None:
            self.line = expr._token[2]
            self.char = expr._token[3]


def _expr_error_message(owner, e):
    """
    Takes an ``EvalError`` or an ``EvalUnitError`` and traces the origins of
    the error in the expression. Returns an error message.

    Arguments:

    ``owner``
        The expression on which eval() or eval_unit() was called.
    ``e``
        The exception that occurred. The method expects the error to have the
        property ``expr`` containing an expression object that triggered the
        error and a property ``err`` that can be converted to a string
        explaining the error.

    """
    def path(root, target, trail=None):
        # Find all name expressions leading up to e.expr
        if trail is None:
            trail = []
        if root == target:
            return trail
        if isinstance(root, Name):
            return path(root.rhs(), target, trail + [root])
        else:
            for child in root:
                r = path(child, target, trail)
                if r is not None:
                    return r
            return None

    # Show there was an error
    out = []
    if isinstance(e, EvalUnitError):
        msg = 'Incompatible units'
        if e.expr._token is not None:
            msg += ' on line ' + str(e.expr._token[2])
        msg += ': ' + str(e.err)
        out.append(msg)
    else:
        out.append(str(e.err))
    out.append('Encountered when evaluating')
    par_str = '  ' + owner.code()
    err_str = e.expr.code()

    # Attempt to locate Name containing error
    # Append list of Names up to that one
    out.append(par_str)
    trail = path(owner, e.expr)
    if trail:
        out.append('Error located at:')
        for i, var in enumerate(trail):
            out.append('  ' * (1 + i) + str(var))
        par_str = str(trail[-1]) + ' = ' + trail[-1].rhs().code()
        out.append(par_str)

    # Indicate location of error
    start = par_str.find(err_str)
    if start >= 0:
        out.append(' ' * start + '~' * len(err_str))

    # Raise new error with added info
    return '\n'.join(out)



#
# SymPy expression writer
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit.formats

try:
    import sympy as sp
except ImportError:
    pass


class SymPyExpressionWriter(myokit.formats.ExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` converts
    Myokit :class:`expressions <myokit.Expression>` to SymPy expressions.

    The returned type is a SymPy object, not a string!
    """
    def __init__(self):
        super().__init__()

        self._flhs = None
        self.set_lhs_function(lambda lhs: str(lhs))

        # Import sympy again, will trigger error if can't be done.
        import sympy
        del sympy

    def set_lhs_function(self, f):
        """
        Sets a naming function, will be called to get the variable name from a
         ``myokit.LhsExpression`` object.

        The argument ``f`` should be a function that takes an ``LhsExpression``
        as input and returns a string.
        """
        self._flhs = f

    def eq(self, eq):
        """
        Converts an equation to a sympy ``Eq`` object.
        """
        return sp.Eq(self.ex(eq.lhs), self.ex(eq.rhs), evaluate=False)

    def ex(self, e):
        """
        Converts an expression to a sympy ``Eq`` object.
        """
        try:
            action = self._op_map[type(e)]
        except KeyError:    # pragma: no cover
            raise ValueError('Unsupported type: ' + str(type(e)))
        return action(e)

    def _ex_name(self, e):
        return sp.Symbol(self._flhs(e))

    def _ex_derivative(self, e):
        # Because a myokit.Derivative doesn't always know what the time
        # variable (or other free variable) is (and because this can change),
        # we handle this using symbols, instead of the sp.Derivative class!
        return sp.Symbol(self._flhs(e))

    def _ex_number(self, e):
        return sp.Float(e.eval())

    def _ex_prefix_plus(self, e):
        return self.ex(e[0])

    def _ex_prefix_minus(self, e):
        return sp.Mul(-1, self.ex(e[0]))

    def _ex_plus(self, e):
        return sp.Add(self.ex(e[0]), self.ex(e[1]))

    def _ex_minus(self, e):
        return sp.Add(self.ex(e[0]), sp.Mul(-1, self.ex(e[1])))

    def _ex_multiply(self, e):
        return sp.Mul(self.ex(e[0]), self.ex(e[1]))

    def _ex_divide(self, e):
        return sp.Mul(self.ex(e[0]), sp.Pow(self.ex(e[1]), -1))

    def _ex_quotient(self, e):
        return self.ex(e[0]) // self.ex(e[1])

    def _ex_remainder(self, e):
        return self.ex(e[0]) % self.ex(e[1])

    def _ex_power(self, e):
        return sp.Pow(self.ex(e[0]), self.ex(e[1]))

    def _ex_sqrt(self, e):
        return sp.Pow(self.ex(e[0]), sp.Rational(1, 2))

    def _ex_sin(self, e):
        return sp.sin(self.ex(e[0]))

    def _ex_cos(self, e):
        return sp.cos(self.ex(e[0]))

    def _ex_tan(self, e):
        return sp.tan(self.ex(e[0]))

    def _ex_asin(self, e):
        return sp.asin(self.ex(e[0]))

    def _ex_acos(self, e):
        return sp.acos(self.ex(e[0]))

    def _ex_atan(self, e):
        return sp.atan(self.ex(e[0]))

    def _ex_exp(self, e):
        return sp.exp(self.ex(e[0]))

    def _ex_log(self, e):
        if len(e) == 1:
            return sp.log(self.ex(e[0]))
        else:
            return sp.log(self.ex(e[0]), self.ex(e[1]))

    def _ex_log10(self, e):
        return sp.log(self.ex(e[0]), 10)

    def _ex_floor(self, e):
        return sp.floor(self.ex(e[0]))

    def _ex_ceil(self, e):
        return sp.ceiling(self.ex(e[0]))

    def _ex_abs(self, e):
        return sp.Abs(self.ex(e[0]))

    def _ex_not(self, e):
        return sp.Not(self.ex(e[0]))

    def _ex_equal(self, e):
        return sp.Eq(self.ex(e[0]), self.ex(e[1]))

    def _ex_not_equal(self, e):
        return sp.Ne(self.ex(e[0]), self.ex(e[1]))

    def _ex_more(self, e):
        return sp.Gt(self.ex(e[0]), self.ex(e[1]))

    def _ex_less(self, e):
        return sp.Lt(self.ex(e[0]), self.ex(e[1]))

    def _ex_more_equal(self, e):
        return sp.Ge(self.ex(e[0]), self.ex(e[1]))

    def _ex_less_equal(self, e):
        return sp.Le(self.ex(e[0]), self.ex(e[1]))

    def _ex_and(self, e):
        return sp.And(self.ex(e[0]), self.ex(e[1]))

    def _ex_or(self, e):
        return sp.Or(self.ex(e[0]), self.ex(e[1]))

    def _ex_if(self, e):
        # Sympy piecewise takes only (expression, condition) pairs
        pairs = []
        pairs.append((self.ex(e._t), self.ex(e._i)))
        pairs.append((self.ex(e._e), True))
        return sp.Piecewise(*pairs)

    def _ex_piecewise(self, e):
        pairs = []
        n = len(e) // 2
        for i in range(0, n):
            pairs.append((self.ex(e._e[i]), self.ex(e._i[i])))
        pairs.append((self.ex(e._e[n]), True))
        return sp.Piecewise(*pairs)




#!/usr/bin/env python3
#
# Tests the CVODES simulation class.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import pickle
import platform
import re
import sys
import unittest

import numpy as np

import myokit

from myokit.tests import (
    CancellingReporter,
    DIR_DATA,
    test_case_pk_model,
    WarningCollector,
)


class SimulationTest(unittest.TestCase):
    """
    Tests the CVODES simulation class.
    """

    @classmethod
    def setUpClass(cls):
        # Test simulation creation.

        m, p, x = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        cls.model = m
        cls.protocol = p
        cls.sim = myokit.Simulation(cls.model, cls.protocol)

    def test_pre(self):
        # Test pre-pacing.

        self.sim.reset()
        self.sim.pre(200)

    def test_simple(self):
        # Test simple run.

        self.sim.reset()
        self.assertEqual(self.sim.time(), 0)
        self.sim.pre(5)
        self.assertEqual(self.sim.time(), 0)
        d = self.sim.run(5)
        self.assertEqual(self.sim.time(), 5)
        self.sim.set_time(0)
        self.assertEqual(self.sim.time(), 0)
        self.assertEqual(type(d), myokit.DataLog)
        self.assertIn('engine.time', d)
        n = len(d['engine.time'])
        for k, v in d.items():
            self.assertEqual(n, len(v))

        # Can't do negative times
        self.assertRaisesRegex(ValueError, 'negative', self.sim.run, -1)

        # Negative log interval is set to zero
        self.sim.reset()
        d1 = self.sim.run(5)
        self.sim.reset()
        d2 = self.sim.run(5, log_interval=-5)
        self.assertEqual(d1.time(), d2.time())

    def test_multiple_protocols(self):
        # Test using multiple protocols

        # Set up a model
        model = myokit.Model()
        c = model.add_component('c')
        t = c.add_variable('t')
        t.set_binding('time')
        t.set_rhs(0)

        a = c.add_variable('a')
        a.set_binding('a')
        a.set_rhs(0)
        b = c.add_variable('b')
        b.set_binding('b')
        b.set_rhs(0)
        y = c.add_variable('y')
        y.promote(1)
        y.set_rhs('-a * y - b * y')

        # Create two overlapping protocols
        pa = myokit.Protocol()
        pa.schedule(level=1, start=0.2, duration=0.5)

        pb = myokit.Protocol()
        pb.schedule(level=2, start=0.5, duration=0.6)

        # Run a simulation with both
        s = myokit.Simulation(model, {'a': pa, 'b': pb})
        s.set_tolerance(1e-8)
        d = s.run(2).npview()
        times = d[t]

        # Check that the changing points are in the log
        self.assertIn(0, times)
        self.assertIn(0.2, times)
        self.assertIn(0.5, times)
        self.assertIn(0.7, times)
        self.assertIn(1.1, times)
        self.assertIn(2, times)

        # Check that a, b, and y have the expected values
        a_up = (times >= 0.2) & (times < 0.7)
        np.testing.assert_array_equal(d[a], np.where(a_up, 1, 0))

        b_up = (times >= 0.5) & (times < 1.1)
        np.testing.assert_array_equal(d[b], np.where(b_up, 2, 0))

        # Check that dy/dt has the expected values
        # dy/dt =
        #   0   from 0 to 0.2
        #   -y  from 0.2 to 0.5
        #   -3y from 0.5 to 0.7
        #   -2y from 0.7 to 1.1
        #   0   after 1.1
        #
        dy_expect = 0 * times
        dy_expect[a_up] -= 1 * d[y][a_up]
        dy_expect[b_up] -= 2 * d[y][b_up]
        np.testing.assert_array_equal(d['dot(c.y)'], dy_expect)

        # Check that y has the expected values
        # The solution for dy/dt = -a*y is y(t) = c * exp(-at), so
        # y =
        #   1
        #   exp(-(t - 0.5))
        #   exp(-0.3) * exp(-3 * (t - 0.5))
        #   exp(-0.9) * exp(-2 * (t - 0.7))
        #   exp(-1.7)
        #
        y_expect = np.ones(times.shape)
        i = (times >= 0.2) & (times < 0.5)
        y_expect[i] = np.exp(-(times[i] - 0.2))
        i = (times >= 0.5) & (times < 0.7)
        y_expect[i] = np.exp(-0.3) * np.exp(-3 * (times[i] - 0.5))
        i = (times >= 0.7) & (times < 1.1)
        y_expect[i] = np.exp(-0.9) * np.exp(-2 * (times[i] - 0.7))
        y_expect[times >= 1.1] = np.exp(-1.7)
        np.testing.assert_array_almost_equal(d[y], y_expect, decimal=3)

        # Test unsetting
        s.set_protocol(None, label='a')
        s.reset()
        d = s.run(2).npview()
        np.testing.assert_array_equal(d[a], np.zeros(d[a].shape))
        times = d[t]
        b_up = (times >= 0.5) & (times < 1.1)
        np.testing.assert_array_equal(d[b], np.where(b_up, 2, 0))
        y_expect = np.ones(times.shape)
        i = (times >= 0.5) & (times < 1.1)
        y_expect[i] = np.exp(-2 * (times[i] - 0.5))
        y_expect[times >= 1.1] = np.exp(-1.2)
        np.testing.assert_array_almost_equal(d[y], y_expect, decimal=3)

        # Test unsetting
        s.set_protocol(None, label='b')
        s.reset()
        d = s.run(2).npview()
        np.testing.assert_array_equal(d[a], np.zeros(d[a].shape))
        np.testing.assert_array_equal(d[a], np.zeros(d[a].shape))
        np.testing.assert_array_equal(d[y], np.ones(d[a].shape))

    def test_mixed_protocols(self):
        # Test using a step and a time series protocol at the same time

        # Set up a model
        model = myokit.Model()
        c = model.add_component('c')
        t = c.add_variable('t')
        t.set_binding('time')
        t.set_rhs(0)
        a = c.add_variable('a')
        a.set_binding('a')
        a.set_rhs(0)
        b = c.add_variable('b')
        b.set_binding('b')
        b.set_rhs(0)
        y = c.add_variable('y')  # Just to log more points
        y.promote(1)
        y.set_rhs('a + b')

        # Set up protocols
        pa = myokit.pacing.blocktrain(level=2, offset=1, duration=2, period=5)
        x = np.linspace(0, 10, 100)
        y = 0.1 + 0.2 * np.sin(x)
        pb = myokit.TimeSeriesProtocol(x, y)

        # Run a simulation with both
        s = myokit.Simulation(model, {'a': pa, 'b': pb})
        s.set_tolerance(1e-8)
        d = s.run(12).npview()
        t = d[t]

        # Start, end, and change points visited exactly
        self.assertIn(0, t)
        self.assertIn(1, t)
        self.assertIn(3, t)
        self.assertIn(6, t)
        self.assertIn(8, t)
        self.assertIn(11, t)
        self.assertIn(12, t)

        # Test a
        a_e = np.zeros(t.shape)
        a_e[(t >= 1) & (t < 3)] = 2
        a_e[(t >= 6) & (t < 8)] = 2
        a_e[(t >= 11) & (t < 13)] = 2
        np.testing.assert_array_equal(d[a], a_e)

        # Test b
        b_e = 0.1 + 0.2 * np.sin(t)
        b_e[t > 10] = y[-1]
        # Note: We don't expect a super good match here, as b_e is an exact
        #       sine value, while the solver will be interpolating y
        np.testing.assert_array_almost_equal(d[b], b_e, decimal=3)

    def test_negative_time(self):
        # Test starting at a negative time

        self.sim.reset()
        self.sim.set_protocol(
            myokit.pacing.blocktrain(duration=1, level=1, period=2))
        self.sim.set_time(-10)
        self.assertEqual(self.sim.time(), -10)
        d = self.sim.run(5, log=['engine.pace']).npview()
        self.assertEqual(self.sim.time(), -5)
        self.assertTrue(np.all(d['engine.pace'] == 0))
        d = self.sim.run(5, log=['engine.pace']).npview()
        self.assertEqual(self.sim.time(), 0)
        self.assertTrue(np.all(d['engine.pace'][:-1] == 0))
        self.assertEqual(d['engine.pace'][-1], 1)
        self.sim.set_protocol(self.protocol)

    def test_no_protocol(self):
        # Test running without a protocol.

        # Check if pace was set to zero (see technical notes).
        self.sim.reset()
        self.sim.pre(50)
        self.sim.set_protocol(None)
        d = self.sim.run(50, log=['engine.pace']).npview()
        self.assertTrue(np.all(d['engine.pace'] == 0))

        # Defined at the start? Then still counts as bound
        self.assertRaisesRegex(
            ValueError, 'not a literal',
            self.sim.set_constant, 'engine.pace', 1)

        # Check that pace is reset to zero when protocol is removed
        x = 0.01    # Note: Must be low to stop sim crashing :D
        self.sim.set_protocol(
            myokit.pacing.blocktrain(duration=1000, level=x, period=1000))
        d = self.sim.run(5, log=['engine.pace']).npview()
        self.assertTrue(np.all(d['engine.pace'] == x))
        self.sim.set_protocol(None)
        d = self.sim.run(5, log=['engine.pace']).npview()
        self.assertTrue(np.all(d['engine.pace'] == 0))
        self.sim.set_protocol(self.protocol)

        # No protocol set from the start? Then "pace" is set anyway for
        # backwards compatibility
        m = self.model.clone()
        m.get('engine.pace').set_rhs(x)
        s = myokit.Simulation(m)
        d = s.run(50).npview()
        self.assertIn('engine.pace', d.keys())
        self.assertTrue(np.all(d['engine.pace'] == 0))

        # But if user explicitly says no labels, then pace isn't treated in a
        # special way.
        s = myokit.Simulation(m, {})
        self.assertRaises(ValueError, s.set_protocol, self.protocol)
        d = s.run(50).npview()
        self.assertNotIn('engine.pace', d.keys())
        #self.assertTrue(np.all(d['engine.pace'] == x))  constant is not logged

    def test_wrong_label_set_pacing(self):
        # Test set_pacing with incorrect label
        self.sim.reset()
        self.sim.pre(50)
        with self.assertRaisesRegex(ValueError, 'Unknown pacing label'):
            self.sim.set_protocol(None, label='does not exist')

    def test_time_series_protocol(self):
        # Test running with a time series protocol

        n = 10
        times = list(range(n))
        values = [0] * n
        values[2:4] = [0.5, 0.5]
        p = myokit.TimeSeriesProtocol(times, values)

        self.sim.set_protocol(p)
        self.sim.reset()
        d = self.sim.run(n, log_interval=1)
        self.assertEqual(list(d.time()), times)
        self.assertEqual(list(d['engine.pace']), values)

        # Unset
        self.sim.set_protocol(None)
        self.sim.reset()
        d = self.sim.run(n, log_interval=1)
        self.assertEqual(list(d['engine.pace']), [0] * n)

        # Reset
        self.sim.set_protocol(p)
        self.sim.reset()
        d = self.sim.run(n, log_interval=1)
        self.assertEqual(list(d.time()), times)
        self.assertEqual(list(d['engine.pace']), values)

        # Unset, replace with original protocol
        self.sim.set_protocol(self.protocol)
        self.sim.reset()
        d = self.sim.run(n, log_interval=1)
        self.assertNotEqual(list(d['engine.pace']), values)
        self.assertNotEqual(list(d['engine.pace']), [0] * n)

        # Deprecated version
        with WarningCollector() as w:
            self.sim.set_fixed_form_protocol(times, values)
        self.assertIn('eprecated', w.text())
        self.sim.reset()
        d = self.sim.run(n, log_interval=1)
        self.assertEqual(list(d.time()), times)
        self.assertEqual(list(d['engine.pace']), values)
        with WarningCollector() as w:
            self.assertRaisesRegex(
                ValueError, 'No times',
                self.sim.set_fixed_form_protocol, values=values)
            self.assertRaisesRegex(
                ValueError, 'No values',
                self.sim.set_fixed_form_protocol, times=times)
            self.sim.set_fixed_form_protocol(None)
        self.sim.reset()
        d = self.sim.run(n, log_interval=1)
        self.assertEqual(list(d['engine.pace']), [0] * n)

        # Reset original protocol
        self.sim.set_protocol(self.protocol)
        self.sim.reset()
        d = self.sim.run(n, log_interval=1)
        self.assertNotEqual(list(d['engine.pace']), values)
        self.assertNotEqual(list(d['engine.pace']), [0] * n)

    def test_in_parts(self):
        # Test running the simulation in parts.

        self.sim.reset()
        # New logs should start with first state, finish with final
        d = self.sim.run(150)
        self.assertEqual(d['engine.time'][0], 0.0)
        self.assertEqual(d['engine.time'][-1], 150.0)
        # Next part should continue at 150, leave where last left off
        e = self.sim.run(50)
        self.assertEqual(d['engine.time'][-1], e['engine.time'][0])
        self.assertEqual(d['membrane.V'][-1], e['membrane.V'][0])
        # Re-used logs shouldn't re-log their first state
        n = len(e['engine.time'])
        e = self.sim.run(50, log=e)
        self.assertNotEqual(e['engine.time'][n - 1], e['engine.time'][n])
        self.assertGreater(e['engine.time'][n], e['engine.time'][n - 1])

    def test_initial_value_expressions(self):
        # Test if initial value expressions are converted to floats

        m = myokit.parse_model('''
            [[model]]
            c.x = 1 + sqrt(3)
            c.y = 1 / c.p
            c.z = 3

            [c]
            t = 0 bind time
            dot(x) = 1
            dot(y) = 2
            dot(z) = 3
            p = log(3)
        ''')
        s = myokit.Simulation(m)
        x = s.state()
        self.assertIsInstance(x[0], float)
        self.assertIsInstance(x[1], float)
        self.assertIsInstance(x[2], float)
        self.assertEqual(x, m.initial_values(True))
        self.assertEqual(x, s.default_state())

    def test_pacing_values_at_event_transitions(self):
        # Tests the value of the pacing signal at event transitions

        # Create a simple model
        m = myokit.Model()
        c = m.add_component('c')
        t = c.add_variable('t')
        t.set_rhs(0)
        t.set_binding('time')
        v = c.add_variable('v')
        v.set_rhs('0')
        v.set_binding('pace')
        x = c.add_variable('x')
        x.set_rhs(0.1)
        x.promote(0)

        # Create step protocol
        p = myokit.Protocol()
        p.schedule(0, 0, 2)
        p.schedule(1, 2, 2)
        p.schedule(2, 4, 4)
        p.schedule(3, 8, 2)

        # Simulate with dynamic logging
        s = myokit.Simulation(m, p)
        d = s.run(p.characteristic_time())
        time = list(d.time())
        value = list(d['c.v'])

        if False:
            for i, t in enumerate(d.time()):
                t = str(np.round(t, 5))
                print(t + ' ' * (10 - len(t)) + str(d['c.v'][i]))

        # Values should be
        #   t   0   1   2   3   4   5   6   7   8   9   10
        #   p   0   0   1   1   2   2   2   2   3   3   0
        self.assertEqual(value[time.index(0.0)], 0)
        self.assertEqual(value[time.index(0.0) + 1], 0)
        self.assertEqual(value[time.index(2.0) - 1], 0)
        self.assertEqual(value[time.index(2.0)], 1)
        self.assertEqual(value[time.index(2.0) + 1], 1)
        self.assertEqual(value[time.index(4.0) - 1], 1)
        self.assertEqual(value[time.index(4.0)], 2)
        self.assertEqual(value[time.index(4.0) + 1], 2)
        self.assertEqual(value[time.index(8.0) - 1], 2)
        self.assertEqual(value[time.index(8.0)], 3)
        self.assertEqual(value[time.index(8.0) + 1], 3)
        self.assertEqual(value[time.index(10.0) - 1], 3)
        self.assertEqual(value[time.index(10.0)], 0)

        # Simulate with fixed logging
        s.reset()
        d = s.run(p.characteristic_time() + 1, log_times=d.time())
        time2 = list(d.time())
        value2 = list(d['c.v'])
        self.assertEqual(time, time2)
        self.assertEqual(value, value2)

    def test_progress_reporter(self):
        # Test running with a progress reporter.

        # Test if it works
        sim = myokit.Simulation(self.model, self.protocol)
        with myokit.tools.capture() as c:
            sim.run(110, progress=myokit.ProgressPrinter())
        c = c.text().splitlines()
        self.assertEqual(len(c), 2)
        p = re.compile(re.escape('[0.0 minutes] 1.9 % done, estimated ') +
                       '[0-9]+' + re.escape(' seconds remaining'))
        self.assertIsNotNone(p.match(c[0]))
        p = re.compile(re.escape('[0.0 minutes] 100.0 % done, estimated ') +
                       '[0-9]+' + re.escape(' seconds remaining'))
        self.assertIsNotNone(p.match(c[1]))

        # Not a progress reporter
        self.assertRaisesRegex(
            ValueError, 'ProgressReporter', self.sim.run, 5, progress=12)

        # Cancel from reporter
        self.assertRaises(
            myokit.SimulationCancelledError, self.sim.run, 1,
            progress=CancellingReporter(0))

    def test_apd_tracking(self):
        # Test the APD / rootfinding method.
        # Tested against offline method in test_datalog.py

        # Test that it works
        self.sim.reset()
        res = self.sim.run(
            1800, log=['engine.time', 'membrane.V'],
            apd_variable='membrane.V', apd_threshold=-70)

        self.assertIsInstance(res, tuple)
        self.assertEqual(len(res), 2)
        d, apds = res
        self.assertIsInstance(d, myokit.DataLog)
        self.assertIsInstance(apds, myokit.DataLog)
        self.assertEqual(len(apds), 2)
        self.assertEqual(apds.length(), 2)
        self.assertIn('start', apds)
        self.assertAlmostEqual(apds['start'][0], 1, places=0)
        self.assertAlmostEqual(apds['start'][1], 1001, places=0)
        self.assertIn('start', apds)
        self.assertAlmostEqual(apds['duration'][0], 383.877194, places=1)
        self.assertAlmostEqual(apds['duration'][1], 378.315915, places=1)

        # Works with variable objects too
        self.sim.reset()
        res = self.sim.run(
            1000, log=['engine.time', 'membrane.V'],
            apd_variable=self.model.get('membrane.V'), apd_threshold=-70)

        # Apd var is not a state
        self.assertRaisesRegex(
            ValueError, 'must be a state',
            self.sim.run, 1000, apd_variable='ina.INa')

        # No apd var given, but threshold provided
        self.assertRaisesRegex(
            ValueError, 'no `apd_variable` specified',
            self.sim.run, 1, apd_threshold=12)

    def test_last_state(self):
        # Returns the last state before an error, or None.

        m = self.model.clone()
        istim = m.get('membrane.i_stim')
        istim.set_rhs('engine.pace / stim_amplitude')
        s = myokit.Simulation(m, self.protocol)
        self.assertIsNone(s.last_state())
        s.run(1)
        self.assertIsNone(s.last_state())
        s.set_constant('membrane.i_stim.stim_amplitude', 0)
        s.reset()
        self.assertRaisesRegex(myokit.SimulationError, "at t = 0", s.run, 5)
        self.assertEqual(len(s.last_state()), len(s.state()))
        self.assertEqual(s.last_state(), s.state())

    def test_last_evaluations_and_steps(self):
        # Test :meth:`Simulation.last_number_of_evaluations()` and
        # :meth:`Simulation.last_number_of_steps()`

        s = myokit.Simulation(self.model, self.protocol)
        self.assertEqual(s.last_number_of_evaluations(), 0)
        self.assertEqual(s.last_number_of_steps(), 0)
        s.run(1)
        self.assertTrue(s.last_number_of_evaluations() > 0)
        self.assertTrue(s.last_number_of_steps() > 0)
        self.assertNotEqual(
            s.last_number_of_evaluations(), s.last_number_of_steps())

    def test_default_state_sensitivites(self):
        # Test :meth:`Simulation.default_state_sensitivies`

        # Create bolus infusion model with linear clearance
        model = myokit.Model()
        comp = model.add_component('myokit')

        amount = comp.add_variable('amount')
        time = comp.add_variable('time')
        dose_rate = comp.add_variable('dose_rate')
        elimination_rate = comp.add_variable('elimination_rate')

        time.set_binding('time')
        dose_rate.set_binding('pace')

        amount.promote(10)
        amount.set_rhs(
            myokit.Minus(
                myokit.Name(dose_rate),
                myokit.Multiply(
                    myokit.Name(elimination_rate),
                    myokit.Name(amount))))
        elimination_rate.set_rhs(myokit.Number(1))
        time.set_rhs(myokit.Number(0))
        dose_rate.set_rhs(myokit.Number(0))

        # Check no sensitivities set
        sim = myokit.Simulation(model)
        self.assertIsNone(sim.default_state_sensitivities())

        # Check for set sensitvities
        sensitivities = (
            ['myokit.amount'],
            ['init(myokit.amount)', 'myokit.elimination_rate'])
        sim = myokit.Simulation(model, sensitivities=sensitivities)
        s = sim.default_state_sensitivities()
        self.assertEqual(len(s), 2)
        self.assertEqual(s[0][0], 1)
        self.assertEqual(s[1][0], 0)

    def test_eval_derivatives(self):
        # Test :meth:`Simulation.eval_derivatives()`.

        self.sim.reset()
        s1 = self.sim.state()
        d1 = self.sim.eval_derivatives()
        self.sim.run(1)
        d2 = self.sim.eval_derivatives()
        self.assertNotEqual(d1, d2)
        self.assertEqual(d1, self.sim.eval_derivatives(s1))
        self.sim.set_state(s1)
        self.assertEqual(d1, self.sim.eval_derivatives())

    def test_eval_derivatives_with_pacing(self):
        # Test :meth:`Simulation.eval_derivatives()`.

        model = myokit.Model()
        c = model.add_component('c')

        a = c.add_variable('a')
        a.set_binding('a')
        a.set_rhs(0)
        b = c.add_variable('b')
        b.set_binding('b')
        b.set_rhs(0)

        y = c.add_variable('y')
        t = c.add_variable('t')
        t.set_binding('time')
        t.set_rhs(0)
        y.promote(1)
        y.set_rhs('- a * y - b * y')

        pa = myokit.Protocol()
        pa.schedule(1, 0, 0.5)

        pb = myokit.Protocol()
        pb.schedule(2, 1.0, 0.5)

        sim = myokit.Simulation(model, {'a': pa, 'b': pb})
        sim.run(1)
        d1 = sim.eval_derivatives(pacing={'a': 0.5, 'b': 0.5})
        d2 = sim.eval_derivatives(pacing={'a': 1.5, 'b': 0.5})
        self.assertNotEqual(d1, d2)
        d1 = sim.eval_derivatives(pacing={'b': 0.5})
        d2 = sim.eval_derivatives(pacing={'a': 0.0, 'b': 0.5})
        self.assertEqual(d1, d2)
        d1 = sim.eval_derivatives()
        d2 = sim.eval_derivatives(pacing={'a': 0.0, 'b': 0.0})
        self.assertEqual(d1, d2)
        d1 = sim.eval_derivatives(pacing={'a': 0.0, 'b': 0.5})
        d2 = sim.eval_derivatives(pacing={'aaaaa': 1.0, 'b': 0.5})
        self.assertEqual(d1, d2)

    def test_sensitivities_initial(self):
        # Test setting initial sensitivity values.

        m = myokit.parse_model('''
            [[model]]
            e.y = 1 + 1.3

            [e]
            t = 0 bind time
            p = 1 / 100
            dot(y) = 2 * p - y
            ''')
        m.validate()

        #TODO: Test results
        s = myokit.Simulation(m, sensitivities=(['e.y'], ['e.p', 'init(e.y)']))

        # Warn if a parameter sensitivity won't be picked up.
        m = myokit.parse_model('''
            [[model]]
            c.x = 1 / c.p

            [c]
            t = 0 bind time
            p = 1 / q
            q = 5
            r = 3
            dot(x) = 2 + r
            ''')
        m.validate()
        s = (['c.x'], ['c.q'])
        self.assertRaisesRegex(
            NotImplementedError, 'respect to parameters used in initial',
            myokit.Simulation, m, sensitivities=s)
        s = (['c.x'], ['c.r'])
        s = myokit.Simulation(m, sensitivities=s)

        # Test bad initial matrix
        self.assertRaisesRegex(
            ValueError, 'None or a list',
            s.run, 10, sensitivities='hello')

    def test_sensitivity_ordering(self):
        # Tests that sensitivities are returned in the correct order

        # Define model
        model = myokit.parse_model(
            """
            [[model]]
            name: one_compartment_pk_model
            # Initial values
            central.drug_amount = 0.1
            dose.drug_amount    = 0.1

            [central]
            dose_rate = 0 bind pace
                in [g/s (1.1574074074074077e-08)]
            dot(drug_amount) = -(
                    size * e.elimination_rate * drug_concentration) + (
                    dose.absorption_rate * dose.drug_amount)
                in [kg (1e-06)]
            drug_concentration = drug_amount / size
                in [g/m^3]
            size = 0.1
                in [L]

            [dose]
            absorption_rate = 0.1
                in [S/F (1.1574074074074077e-05)]
            dot(drug_amount) = (-absorption_rate * drug_amount
                                + central.dose_rate)
                in [kg (1e-06)]

            [e]
            elimination_rate = 0.1
                in [S/F (1.1574074074074077e-05)]
            time = 0 bind time
                in [s (86400)]
            """
        )

        # Select all states and possible independents
        var = ['central.drug_amount', 'dose.drug_amount']
        par = [
            'central.size',
            'init(central.drug_amount)',
            'dose.absorption_rate',
            'init(dose.drug_amount)',
            'e.elimination_rate',
        ]

        # Run and store
        sim = myokit.Simulation(model, sensitivities=(var, par))
        _, s1 = sim.run(6, log=myokit.LOG_NONE, log_interval=2)
        s1 = np.array(s1)

        # Now use reverse order
        var.reverse()
        par.reverse()

        # Run and store
        sim = myokit.Simulation(model, sensitivities=(var, par))
        _, s2 = sim.run(6, log=myokit.LOG_NONE, log_interval=2)
        s2 = np.array(s2)

        # Reverse results, to get back original order
        s2 = s2[:, ::-1, ::-1]
        self.assertTrue(np.all(s1 == s2))

    def test_sensitivities_bolus_infusion(self):
        # Test :meth:`Simulation.run()` accuracy by comparing to
        # analytic solution in a PKPD bolus infusion model.

        # Create bolus infusion model with linear clearance
        model = myokit.Model()
        comp = model.add_component('myokit')

        amount = comp.add_variable('amount')
        time = comp.add_variable('time')
        dose_rate = comp.add_variable('dose_rate')
        elimination_rate = comp.add_variable('elimination_rate')

        time.set_binding('time')
        dose_rate.set_binding('pace')

        amount.promote(10)
        amount.set_rhs(
            myokit.Minus(
                myokit.Name(dose_rate),
                myokit.Multiply(
                    myokit.Name(elimination_rate),
                    myokit.Name(amount))))
        elimination_rate.set_rhs(myokit.Number(1))
        time.set_rhs(myokit.Number(0))
        dose_rate.set_rhs(myokit.Number(0))

        # Create protocol
        amount = 5
        duration = 0.001
        protocol = myokit.pacing.blocktrain(
            1, duration, offset=0, level=amount / duration, limit=0)

        # Set sensitivies
        sensitivities = (
            ['myokit.amount'],
            ['init(myokit.amount)', 'myokit.elimination_rate'])

        sim = myokit.Simulation(model, protocol, sensitivities)

        # Set tolerance to be controlled by abs_tolerance for
        # easier comparison
        #TODO This rel_tol is insanely high
        sim.set_tolerance(abs_tol=1e-8, rel_tol=1e-30)

        # Solve problem analytically
        parameters = [10, 1, amount / duration, duration]
        times = np.linspace(0.1, 10, 13)
        ref_sol, ref_partials = test_case_pk_model(parameters, times)

        # Solve problem with simulator
        sol, partials = sim.run(
            11, log_times=times, log=['myokit.amount'])
        sol = np.array(sol['myokit.amount'])
        partials = np.array(partials).squeeze()

        # Check state
        self.assertAlmostEqual(sol[0], ref_sol[0], 6)
        self.assertAlmostEqual(sol[1], ref_sol[1], 6)
        self.assertAlmostEqual(sol[2], ref_sol[2], 6)
        self.assertAlmostEqual(sol[3], ref_sol[3], 6)
        self.assertAlmostEqual(sol[4], ref_sol[4], 6)
        self.assertAlmostEqual(sol[5], ref_sol[5], 6)
        self.assertAlmostEqual(sol[6], ref_sol[6], 6)
        self.assertAlmostEqual(sol[7], ref_sol[7], 6)
        self.assertAlmostEqual(sol[8], ref_sol[8], 6)
        self.assertAlmostEqual(sol[9], ref_sol[9], 6)
        self.assertAlmostEqual(sol[10], ref_sol[10], 6)
        self.assertAlmostEqual(sol[11], ref_sol[11], 6)
        self.assertAlmostEqual(sol[12], ref_sol[12], 6)

        # Check partials
        self.assertAlmostEqual(partials[0, 0], ref_partials[0, 0], 6)
        self.assertAlmostEqual(partials[1, 0], ref_partials[0, 1], 6)
        self.assertAlmostEqual(partials[2, 0], ref_partials[0, 2], 6)
        self.assertAlmostEqual(partials[3, 0], ref_partials[0, 3], 6)
        self.assertAlmostEqual(partials[4, 0], ref_partials[0, 4], 6)
        self.assertAlmostEqual(partials[5, 0], ref_partials[0, 5], 6)
        self.assertAlmostEqual(partials[6, 0], ref_partials[0, 6], 6)
        self.assertAlmostEqual(partials[7, 0], ref_partials[0, 7], 6)
        self.assertAlmostEqual(partials[8, 0], ref_partials[0, 8], 6)
        self.assertAlmostEqual(partials[9, 0], ref_partials[0, 9], 6)
        self.assertAlmostEqual(partials[10, 0], ref_partials[0, 10], 6)
        self.assertAlmostEqual(partials[11, 0], ref_partials[0, 11], 6)
        self.assertAlmostEqual(partials[12, 0], ref_partials[0, 12], 6)

        # Return apds
        result = sim.run(1, apd_variable='myokit.amount', apd_threshold=0.3)
        self.assertEqual(len(result), 3)

    def test_sensitivities_with_derivatives(self):
        # Test various inputs to the `sensitivities` argument are handled
        # correctly, including sensitivites of derivatives and intermediary
        # variables depending on derivatives.

        # Note: passing the wrong values into the sensitivities constructor arg
        # is tested in the CModel tests.

        m = myokit.parse_model('''
            [[model]]
            e.x = 1.2
            e.y = 2.3

            [e]
            t = 0 bind time
            p = 1 / 100
            q = 2
            r = 3
            dot(x) = p * q
            dot(y) = 2 * p - y
            fx = x^2
            fy = r + p^2 + dot(y)
            ''')
        m.validate()

        x0 = m.get('e.x').initial_value(True)
        y0 = m.get('e.y').initial_value(True)
        p = m.get('e.p').eval()
        q = m.get('e.q').eval()
        r = m.get('e.r').eval()

        # Dependents is a list of y in dy/dx.
        # Must refer to state, derivative, or intermediary, given as:
        #  - Variable
        #  - Name
        #  - Derivative
        #  - "qname"
        #  - "dot(qname)"
        dependents = [
            'e.x',                  # qname
            'e.y',                  # qname
            'dot(e.x)',             # dot(qname)
            m.get('e.y').lhs(),     # Derivative
            m.get('e.fx'),          # Variable
            m.get('e.fy').lhs(),    # Name
        ]

        # Independents is a list of x in dy/dx
        # Must refer to literals or initial values, given as:
        #  - Variable
        #  - Name
        #  - InitialValue
        #  - "qname"
        #  - "init(qname)"
        independents = [
            'e.p',                  # qname
            m.get('e.q'),           # qname
            m.get('e.r').lhs(),     # Name
            'init(e.x)',            # init(qname)
            myokit.InitialValue(myokit.Name(m.get('e.y'))),  # InitialValue
        ]

        # Run, get output
        s = myokit.Simulation(m, sensitivities=(dependents, independents))
        s.set_tolerance(1e-9, 1e-9)
        d, e = s.run(8)
        d, e = d.npview(), np.array(e)
        t = d.time()

        # Check solution
        self.assertTrue(
            np.allclose(d['e.x'], x0 + p * q * t))
        self.assertTrue(np.allclose(
            d['e.y'], 2 * p + (y0 - 2 * p) * np.exp(-t)))
        self.assertTrue(np.allclose(
            d['e.fx'], (p * q * t)**2 + 2 * p * q * t * x0 + x0**2))
        self.assertTrue(np.allclose(
            d['e.fy'], r + p**2 + (2 * p - y0) * np.exp(-t)))
        self.assertTrue(np.allclose(
            d['dot(e.x)'], p * q))
        self.assertTrue(np.allclose(
            d['dot(e.y)'], (2 * p - y0) * np.exp(-t)))

        # Sensitivities of  x  to (p, q, r, x0, y0)
        self.assertTrue(np.allclose(e[:, 0, 0], q * t))
        self.assertTrue(np.allclose(e[:, 0, 1], p * t))
        self.assertTrue(np.allclose(e[:, 0, 2], 0))
        self.assertTrue(np.allclose(e[:, 0, 3], 1))
        self.assertTrue(np.allclose(e[:, 0, 4], 0))

        # Sensitivities of  y  to (p, q, r, x0, y0)
        self.assertTrue(np.allclose(e[:, 1, 0], 2 - 2 * np.exp(-t)))
        self.assertTrue(np.allclose(e[:, 1, 1], 0))
        self.assertTrue(np.allclose(e[:, 1, 2], 0))
        self.assertTrue(np.allclose(e[:, 1, 3], 0))
        self.assertTrue(np.allclose(e[:, 1, 4], np.exp(-t)))

        # Sensitivities of  dot(x)  to (p, q, r, x0, y0)
        self.assertTrue(np.allclose(e[:, 2, 0], q))
        self.assertTrue(np.allclose(e[:, 2, 1], p))
        self.assertTrue(np.allclose(e[:, 2, 2], 0))
        self.assertTrue(np.allclose(e[:, 2, 3], 0))
        self.assertTrue(np.allclose(e[:, 2, 4], 0))

        # Sensitivities of  dot(y)  to (p, q, r, x0, y0)
        self.assertTrue(np.allclose(e[:, 3, 0], 2 * np.exp(-t)))
        self.assertTrue(np.allclose(e[:, 3, 1], 0))
        self.assertTrue(np.allclose(e[:, 3, 2], 0))
        self.assertTrue(np.allclose(e[:, 3, 3], 0))
        self.assertTrue(np.allclose(e[:, 3, 4], -np.exp(-t)))

        # Sensitivities of  fx  to (p, q, r, x0, y0)
        self.assertTrue(np.allclose(
            e[:, 4, 0], 2 * p * (q * t)**2 + 2 * q * t * x0))
        self.assertTrue(np.allclose(
            e[:, 4, 1], 2 * q * (p * t)**2 + 2 * p * t * x0))
        self.assertTrue(np.allclose(
            e[:, 4, 2], 0))
        self.assertTrue(np.allclose(
            e[:, 4, 3], 2 * p * q * t + 2 * x0))
        self.assertTrue(np.allclose(
            e[:, 4, 4], 0))

        # Sensitivities of  fy  to (p, q, r, x0, y0)
        self.assertTrue(np.allclose(e[:, 5, 0], 2 * p + 2 * np.exp(-t)))
        self.assertTrue(np.allclose(e[:, 5, 1], 0))
        self.assertTrue(np.allclose(e[:, 5, 2], 1))
        self.assertTrue(np.allclose(e[:, 5, 3], 0))
        self.assertTrue(np.allclose(e[:, 5, 4], -np.exp(-t)))

    def test_set_tolerance(self):
        # Test :meth:`Simulation.set_tolerance()`.

        self.assertRaisesRegex(
            ValueError, 'Absolute', self.sim.set_tolerance, abs_tol=0)
        self.assertRaisesRegex(
            ValueError, 'Relative', self.sim.set_tolerance, rel_tol=0)
        self.sim.set_tolerance(1e-6, 1e-4)

    def test_set_step_size(self):
        # Test :meth:`Simulation.set_min_step_size()` and
        # :meth:`Simulation.set_max_step_size()`.

        # Minimum: set, unset, allow negative value to unset
        self.sim.set_min_step_size(0.1)
        self.sim.set_min_step_size(None)
        self.sim.set_min_step_size(-1)

        # Same for max
        self.sim.set_max_step_size(0.1)
        self.sim.set_max_step_size(None)
        self.sim.set_max_step_size(-1)

    def test_set_state(self):
        # Test :meth:`Simulation.set_state()` and
        # :meth:`Simulation.set_default_state()`.

        # Get state and default state, both different from current
        state = self.sim.state()
        state[0] += 1
        default_state = self.sim.default_state()
        default_state[1] += 1
        if state == default_state:
            default_state[0] += 2

        self.assertNotEqual(self.sim.state(), state)
        self.assertNotEqual(self.sim.default_state(), default_state)

        self.sim.set_state(state)
        self.sim.set_default_state(default_state)

        self.assertEqual(self.sim.state(), state)
        self.assertEqual(self.sim.default_state(), default_state)

    def test_set_constant(self):
        # Test :meth:`Simulation.set_constant()`.

        # Literal
        self.sim.set_constant('cell.Na_i', 11)
        self.sim.set_constant(self.model.get('cell.Na_i'), 11)
        self.assertRaises(KeyError, self.sim.set_constant, 'cell.Bert', 11)

        # Parameter (needs sensitivies set)
        m, p, x = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        sim = myokit.Simulation(m, p, (['ib.Ib'], ['ib.gb']))
        sim.set_constant('ib.gb', 20)

        # Calculated constant
        self.assertRaisesRegex(
            ValueError, 'not a literal', self.sim.set_constant, 'ina.ENa', 11)

    def test_short_runs(self):
        # Test for simulations run a very short time

        # Run for 1 unit (OK)
        self.sim.reset()
        self.sim.run(1)

        # Test running for 0 units doesn't affect state
        x0 = self.sim.state()
        self.sim.run(0)
        self.assertEqual(x0, self.sim.state())
        self.sim.run(0)
        self.assertEqual(x0, self.sim.state())

        # Test running between indistinguishable times doesn't affect state
        t = self.sim.time()
        d = 0.5 * sys.float_info.epsilon
        self.assertEqual(t, t + d)
        self.sim.run(d)

        # Test running between only just distinguishable times is fine
        self.sim.reset()
        self.sim.run(1)
        t = self.sim.time()
        d = 3 * sys.float_info.epsilon
        self.assertNotEqual(t, t + d)
        self.sim.run(d)

        # Test running between barely distinguishable times raises CVODE error.
        t = self.sim.time()
        d = 2 * sys.float_info.epsilon
        self.assertNotEqual(t, t + d)
        self.assertRaisesRegex(
            myokit.SimulationError, 'CV_TOO_CLOSE', self.sim.run, d)

    def test_simulation_error_1(self):
        # Test for simulation error detection: massive stimulus.

        # Silly protocol
        p = myokit.Protocol()
        p.schedule(level=1000, start=1, duration=1)
        self.sim.reset()
        self.sim.set_protocol(p)
        self.assertRaisesRegex(
            myokit.SimulationError, 'numerical error', self.sim.run, 10)
        self.sim.set_protocol(self.protocol)

    @unittest.skipIf(platform.system() != 'Linux', 'Cvode error tests')
    def test_simulation_error_2(self):
        # Test for simulation error detection: failure occurred too often.

        # Cvode error (test failure occurred too many times)
        m = self.model.clone()
        v = m.get('membrane.V')
        v.set_rhs(myokit.Multiply(v.rhs(), myokit.Number(1e18)))
        s = myokit.Simulation(m, self.protocol)
        with WarningCollector():
            self.assertRaisesRegex(
                myokit.SimulationError, 'numerical error', s.run, 5000)

    def test_cvode_simulation_with_zero_states(self):
        # Tests running cvode simulations on models with no ODEs

        # Create a model without states
        m1 = myokit.Model()
        c = m1.add_component('c')
        t = c.add_variable('t')
        t.set_rhs(0)
        t.set_binding('time')
        v = c.add_variable('v')
        v.set_rhs('0')
        v.set_binding('pace')
        w = c.add_variable('w')
        w.set_rhs('2 * v')

        # Create a model with a state
        m2 = m1.clone()
        z = m2.get('c').add_variable('z')
        z.set_rhs(0.1)
        z.promote(0)

        # Test without protocol and dynamic logging
        #myokit.DEBUG_WG = True
        s1 = myokit.Simulation(m1)
        d1 = s1.run(5)
        self.assertEqual(len(d1.time()), 2)
        self.assertEqual(list(d1.time()), [0, 5])
        self.assertEqual(list(d1['c.w']), [0, 0])
        s2 = myokit.Simulation(m2)
        d2 = s2.run(6, log_times=d1.time())
        self.assertEqual(d1.time(), d2.time())
        self.assertEqual(d1['c.w'], d2['c.w'])

        # Test with a protocol and dynamic logging
        p = myokit.Protocol()
        p.schedule(0, 0, 2)
        p.schedule(1, 2, 2)
        p.schedule(2, 4, 4)
        p.schedule(3, 8, 2)
        s1.reset()
        s1.set_protocol(p)
        d1 = s1.run(p.characteristic_time())
        self.assertEqual(len(d1.time()), 5)
        self.assertEqual(list(d1.time()), [0, 2, 4, 8, 10])
        self.assertEqual(list(d1['c.w']), [0, 2, 4, 6, 0])
        s2.reset()
        s2.set_protocol(p)
        d2 = s2.run(p.characteristic_time() + 1, log_times=d1.time())
        self.assertEqual(d1.time(), d2.time())
        self.assertEqual(d1['c.w'], d2['c.w'])

        # Test with fixed logging times
        s1.reset()
        d1 = s1.run(p.characteristic_time() + 1, log_times=d1['c.t'])
        self.assertEqual(list(d1.time()), [0, 2, 4, 8, 10])
        self.assertEqual(list(d1['c.w']), [0, 2, 4, 6, 0])
        s2.reset()
        d2 = s2.run(p.characteristic_time() + 1, log_times=d1.time())
        self.assertEqual(d1.time(), d2.time())
        self.assertEqual(d1['c.w'], d2['c.w'])

        # Test appending to log
        s1.reset()
        d1 = s1.run(5)
        d1 = s1.run(5, log=d1)
        self.assertEqual(list(d1.time()), [0, 2, 4, 5, 8, 10])
        self.assertEqual(list(d1['c.w']), [0, 2, 4, 4, 6, 0])

        # Test with a log interval
        s1.reset()
        d1 = s1.run(11, log_interval=1)
        self.assertEqual(list(d1.time()), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        self.assertEqual(list(d1['c.w']), [0, 0, 2, 2, 4, 4, 4, 4, 6, 6, 0])

    def test_pickling(self):
        # Test pickling a simulation

        # Test with myokit.Protocol and sensitivities
        m, p, _ = myokit.load('example')
        s1 = myokit.Simulation(m, p, [
            ['ina.INa', 'ica.ICa', 'membrane.V'],
            ['ina.gNa', 'init(membrane.V)']
        ])
        s1.pre(123)
        # ...and with simulation properties
        s1.set_tolerance(1e-8, 1e-8)
        s1.set_min_step_size(1e-4)
        s1.set_max_step_size(0.1)
        # ...and changed properties
        s1.set_constant('membrane.C', 1.1)
        s_bytes = pickle.dumps(s1)
        s2 = pickle.loads(s_bytes)
        self.assertEqual(s1.time(), s2.time())
        self.assertEqual(s1.state(), s2.state())
        self.assertEqual(s1.default_state(), s2.default_state())
        d1, e1 = s1.run(123, log=myokit.LOG_NONE)
        d2, e2 = s2.run(123, log=myokit.LOG_NONE)
        self.assertEqual(s1.time(), s2.time())
        self.assertEqual(s1.state(), s2.state())
        self.assertTrue(np.all(e1 == e2))

    def test_sim_stats(self):
        # Test extraction of simulation statistics
        m, p, _ = myokit.load('example')
        rt = m['engine'].add_variable('realtime')
        rt.set_rhs(0)
        rt.set_binding('realtime')
        ev = m['engine'].add_variable('evaluations')
        ev.set_rhs(0)
        ev.set_binding('evaluations')
        s = myokit.Simulation(m, p)
        d = s.run(100, log=myokit.LOG_BOUND).npview()

        self.assertIn('engine.realtime', d)
        self.assertIn('engine.evaluations', d)
        rt, ev = d['engine.realtime'], d['engine.evaluations']
        self.assertEqual(len(d.time()), len(rt))
        self.assertEqual(len(d.time()), len(ev))
        self.assertTrue(np.all(rt >= 0))
        self.assertTrue(np.all(ev >= 0))
        self.assertTrue(np.all(rt[1:] >= rt[:-1]))
        self.assertTrue(np.all(ev[1:] >= ev[:-1]))

    def test_derivatives(self):
        # Tests logging of derivatives by comparing with a finite difference
        # approximation

        # Run past the upstroke, where finite diff approx is worst
        self.sim.reset()
        self.sim.run(52)

        # Now run logged part
        d = self.sim.run(600).npview()
        if False:
            import matplotlib.pyplot as plt
            plt.figure()
            ax = plt.subplot(3, 1, 1)
            ax.plot(d.time(), d['membrane.V'])
            ax = plt.subplot(3, 1, 2)
            ax.plot(d.time(), d['dot(membrane.V)'])

        # Get central difference approximation
        t = d.time()
        v = d['membrane.V']
        dv = (v[2:] - v[:-2]) / (t[2:] - t[:-2])
        e = d['dot(membrane.V)'][1:-1] - dv
        if False:
            t = t[1:-1]
            ax.plot(t, dv, '--')
            ax = plt.subplot(3, 1, 3)
            ax.plot(t, e)
            print(np.max(np.abs(e)))

        # Compare
        self.assertLess(np.max(np.abs(e)), 0.1)


if __name__ == '__main__':
    unittest.main()
